import json

import pytest
from py._path.local import LocalPath  # type:ignore[import]
from pydantic import ValidationError

from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.com.adresse import Adresse
from bo4e.enum.anrede import Anrede
from bo4e.enum.botyp import BoTyp
from bo4e.enum.geschaeftspartnerrolle import Geschaeftspartnerrolle
from bo4e.enum.kontaktart import Kontaktart
from bo4e.enum.landescode import Landescode


class TestGeschaeftspartner:
    @pytest.mark.datafiles("./tests/test_data/test_data_adresse/test_data_adresse_only_required_fields.json")
    def test_serializable(self, datafiles: LocalPath) -> None:
        with open(datafiles / "test_data_adresse_only_required_fields.json", encoding="utf-8") as json_file:
            address_test_data = json.load(json_file)

        gp = Geschaeftspartner(
            anrede=Anrede.FRAU,
            name1="von Sinnen",
            name2="Helga",
            name3=None,
            gewerbekennzeichnung=True,
            hrnummer="HRB 254466",
            amtsgericht="Amtsgericht München",
            kontaktweg=[Kontaktart.E_MAIL],
            umsatzsteuer_id="DE267311963",
            glaeubiger_id="DE98ZZZ09999999999",
            e_mail_adresse="test@bo4e.de",
            website="bo4e.de",
            geschaeftspartnerrolle=[Geschaeftspartnerrolle.DIENSTLEISTER],
            partneradresse=Adresse(
                postleitzahl=address_test_data["postleitzahl"],
                ort=address_test_data["ort"],
                strasse=address_test_data["strasse"],
                hausnummer=address_test_data["hausnummer"],
            ),
        )

        # test default value for bo_typ in Geschaeftspartner
        assert gp.bo_typ == BoTyp.GESCHAEFTSPARTNER

        gp_json = gp.json(by_alias=True, ensure_ascii=False)

        assert "Helga" in gp_json

        gp_deserialized = Geschaeftspartner.parse_raw(gp_json)

        assert gp_deserialized.bo_typ == gp.bo_typ
        assert type(gp_deserialized.partneradresse) == Adresse

    def test_optional_attribute_partneradresse(self) -> None:
        """
        The BO4E standard does not yet define whether the partneradresse is mandatory or not.
        We will set this as an optional argument until the standard is clear about it.

        This test checks whether the Geschaeftspartner can also be initialized without a partneradresse.
        """

        gp = Geschaeftspartner(
            anrede=Anrede.FRAU,
            name1="von Sinnen",
            name2="Helga",
            name3=None,
            gewerbekennzeichnung=True,
            hrnummer="HRB 254466",
            amtsgericht="Amtsgericht München",
            kontaktweg=[Kontaktart.E_MAIL],
            umsatzsteuer_id="DE267311963",
            glaeubiger_id="DE98ZZZ09999999999",
            e_mail_adresse="test@bo4e.de",
            website="bo4e.de",
            geschaeftspartnerrolle=[Geschaeftspartnerrolle.DIENSTLEISTER],
        )

        gp_json = gp.json(by_alias=True, ensure_ascii=False)
        gp_deserialized = Geschaeftspartner.parse_raw(gp_json)

        assert gp_deserialized.partneradresse is None

    def test_list_validation_of_geschaeftspartnerrolle(self) -> None:
        """
        Tests that if the geschaeftspartnerrolle of Geschaeftspartner is not a list, an error is raised.

        The attribute geschaeftspartnerrolle of Geschaeftspartner must be a list type.
        Therefore the list validator checks the type of geschaeftspartnerrolle
        during the initialization of Geschaeftspartner.
        """

        with pytest.raises(ValidationError) as excinfo:
            _ = Geschaeftspartner(
                anrede=Anrede.FRAU,
                name1="von Sinnen",
                name2="Helga",
                name3=None,
                gewerbekennzeichnung=True,
                hrnummer="HRB 254466",
                amtsgericht="Amtsgericht München",
                kontaktweg=[Kontaktart.E_MAIL],
                umsatzsteuer_id="DE267311963",
                glaeubiger_id="DE98ZZZ09999999999",
                e_mail_adresse="test@bo4e.de",
                website="bo4e.de",
                geschaeftspartnerrolle=Geschaeftspartnerrolle.DIENSTLEISTER,  # type: ignore[arg-type]
                partneradresse=Adresse(
                    postleitzahl="01069",
                    ort="Dresden",
                    strasse="Goethestraße",
                    hausnummer="1",
                ),
            )

        assert "1 validation error" in str(excinfo.value)
        assert "value is not a valid list" in str(excinfo.value)

    def test_serialization_of_non_german_address(self) -> None:
        """
        Test that partneradresses with a Landescode!=DE (default) are (de)serialized correctly.
        """
        gp = Geschaeftspartner(
            anrede=Anrede.FRAU,
            name1="Kurz",
            name2="Sebastian",
            name3=None,
            gewerbekennzeichnung=True,
            hrnummer="HRB 254466",
            amtsgericht="Amtsgericht Ibiza",
            kontaktweg=[Kontaktart.E_MAIL],
            umsatzsteuer_id="AT12345",
            geschaeftspartnerrolle=[Geschaeftspartnerrolle.DIENSTLEISTER],
            partneradresse=Adresse(
                postleitzahl="1014", ort="Wien 1", strasse="Ballhausplatz", hausnummer="2", landescode=Landescode.AT  # type: ignore[attr-defined]
            ),
        )
        gp_json = gp.json(by_alias=True, ensure_ascii=False)
        gp_deserialized = Geschaeftspartner.parse_raw(gp_json)
        assert gp_deserialized.partneradresse.landescode == Landescode.AT  # type: ignore[attr-defined, union-attr]



from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.kriteriumwert import KriteriumWert
from bo4e.enum.tarifregionskriterium import Tarifregionskriterium
from tests.serialization_helper import assert_serialization_roundtrip


class TestKriteriumWert:
    @pytest.mark.parametrize(
        "kriteriumwert, expected_json_dict",
        [
            pytest.param(
                KriteriumWert(
                    kriterium=Tarifregionskriterium.ORT,
                    wert="Grünwald",
                ),
                {"kriterium": Tarifregionskriterium.ORT, "wert": "Grünwald"},
            ),
        ],
    )
    def test_kriteriumwert_serialization_roundtrip(
        self, kriteriumwert: KriteriumWert, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of KriteriumWert with minimal attributes.
        """
        assert_serialization_roundtrip(kriteriumwert, expected_json_dict)

    def test_kriteriumwert_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = KriteriumWert()  # type: ignore[call-arg]
        assert "2 validation errors" in str(excinfo.value)



"""
Contains Zeitreihenwertkompakt class
and corresponding marshmallow schema for de-/serialization
"""
from decimal import Decimal
from typing import Optional

from bo4e.com.com import COM
from bo4e.enum.messwertstatus import Messwertstatus
from bo4e.enum.messwertstatuszusatz import Messwertstatuszusatz

# pylint: disable=too-few-public-methods


class Zeitreihenwertkompakt(COM):
    """
    Abbildung eines kompakten Zeitreihenwertes in dem ausschliesslich der Wert und Statusinformationen stehen.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Zeitreihenwertkompakt.svg" type="image/svg+xml"></object>

    .. HINT::
        `Zeitreihenwertkompakt JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Zeitreihenwertkompakt.json>`_

    """

    # required attributes
    wert: Decimal  #: Der im Zeitintervall gültige Wert.

    # optional attributes

    #: Der Status gibt an, wie der Wert zu interpretieren ist, z.B. in Berechnungen.
    status: Optional[Messwertstatus] = None

    #: Eine Zusatzinformation zum Status, beispielsweise ein Grund für einen fehlenden Wert.
    statuszusatz: Optional[Messwertstatuszusatz] = None



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Mengeneinheit(StrEnum):
    """
    Einheit: Messgrößen, die per Messung oder Vorgabe ermittelt werden können.
    """

    W = "W"  #: Watt
    WH = "WH"  #: Wattstunde
    KW = "KW"  #: Kilowatt
    KWH = "KWH"  #: Kilowattstunde
    KVARH = "KVARH"  #: Kilovarstunde
    MW = "MW"  #: Megawatt
    MWH = "MWH"  #: Megawattstunde
    STUECK = "STUECK"  #: Stückzahl
    KUBIKMETER = "KUBIKMETER"  #: Kubikmeter (Gas)
    STUNDE = "STUNDE"  #: Stunde
    TAG = "TAG"  #: Tage
    MONAT = "MONAT"  #: Monat
    JAHR = "JAHR"  #: Jahr
    PROZENT = "PROZENT"  #: Prozent



from bo4e.com.marktgebietinfo import MarktgebietInfo


class TestMarktgebietinfo:
    def test_serialization(self) -> None:
        mgi = MarktgebietInfo(marktgebiet="Gaspool", marktgebietcode="37Z701133MH0000B")

        json_string = mgi.json(by_alias=True, ensure_ascii=False)

        assert "marktgebiet" in json_string, "No camel case serialization"

        deserialized_mgi: MarktgebietInfo = MarktgebietInfo.parse_raw(json_string)

        assert mgi == deserialized_mgi



"""
Contains PositionsAufAbschlag and corresponding marshmallow schema for de-/serialization
"""
from decimal import Decimal
from typing import Optional

from bo4e.com.com import COM

# pylint: disable=too-few-public-methods
from bo4e.enum.aufabschlagstyp import AufAbschlagstyp
from bo4e.enum.waehrungseinheit import Waehrungseinheit


class PositionsAufAbschlag(COM):
    """
    Differenzierung der zu betrachtenden Produkte anhand der preiserhöhenden (Aufschlag)
    bzw. preisvermindernden (Abschlag) Zusatzvereinbarungen,
    die individuell zu einem neuen oder bestehenden Liefervertrag abgeschlossen werden können.
    Es können mehrere Auf-/Abschläge gleichzeitig ausgewählt werden.

    .. raw:: html

        <object data="../_static/images/bo4e/com/PositionsAufAbschlag.svg" type="image/svg+xml"></object>

    .. HINT::
        `PositionsAufAbschlag JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/PositionsAufAbschlag.json>`_

    """

    # required attributes
    #: Bezeichnung des Auf-/Abschlags
    bezeichnung: str
    #: Beschreibung zum Auf-/Abschlag
    beschreibung: Optional[str] = None
    #: Typ des AufAbschlages
    auf_abschlagstyp: AufAbschlagstyp
    #: Höhe des Auf-/Abschlages
    auf_abschlagswert: Decimal
    #: Einheit, in der der Auf-/Abschlag angegeben ist (z.B. ct/kWh).
    auf_abschlagswaehrung: Waehrungseinheit



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Messgroesse(StrEnum):
    """
    Gibt die physikalische Größe an, die gemessen wurde.
    """

    STROM = "STROM"  #: Strom
    SPANNUNG = "SPANNUNG"  #: Spannung
    WIRKLEISTUNG = "WIRKLEISTUNG"  #: Wirkleistung
    BLINDLEISTUNG = "BLINDLEISTUNG"  #: Blindleistung
    DRUCK = "DRUCK"  #: Druck
    LASTGANG = "LASTGANG"  #: Lastgang
    LASTPROFIL = "LASTPROFIL"  #: Lastprofil
    TEMPERATUR = "TEMPERATUR"  #: Temperatur
    ZZAHL = "ZZAHL"  #: Zustandszahl
    BRENNWERT = "BRENNWERT"  #: Brennwert
    GRADTZAGSZAHLEN = "GRADTZAGSZAHLEN"  #: Gradtagszahlen
    VOLUMENSTROM = "VOLUMENSTROM"  #: Volumenstrom
    PREISE = "PREISE"  #: Preise



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Vertragsform(StrEnum):
    """
    Aufzählung der Möglichkeiten zu Vertragsformen in Ausschreibungen.
    """

    ONLINE = "ONLINE"  #: Online
    DIREKT = "DIREKT"  #: Direkt
    FAX = "FAX"  #: Auftragsfax



"""
Contains Tarifberechnungsparameter class
and corresponding marshmallow schema for de-/serialization
"""
from decimal import Decimal
from typing import List, Optional

from bo4e.com.com import COM
from bo4e.com.preis import Preis
from bo4e.com.tarifpreis import Tarifpreis
from bo4e.enum.messpreistyp import Messpreistyp
from bo4e.enum.tarifkalkulationsmethode import Tarifkalkulationsmethode

# yes. there is no description in the official docs.
# https://github.com/Hochfrequenz/BO4E-python/issues/328

# pylint: disable=too-few-public-methods, empty-docstring, too-many-instance-attributes


class Tarifberechnungsparameter(COM):
    """
    In dieser Komponente sind die Berechnungsparameter für die Ermittlung der Tarifkosten zusammengefasst.
    .. raw:: html

        <object data="../_static/images/bo4e/com/Tarifberechnungsparameter.svg" type="image/svg+xml"></object>

    .. HINT::
        `Tarifberechnungsparameter JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Tarifberechnungsparameter.json>`_

    """

    # there are no required attributes
    # optional attributes

    #: Gibt an, wie die Einzelpreise des Tarifes zu verarbeiten sind
    berechnungsmethode: Optional[Tarifkalkulationsmethode] = None
    #: True, falls der Messpreis im Grundpreis (GP) enthalten ist
    messpreis_in_gp_enthalten: Optional[bool] = None

    messpreis_beruecksichtigen: Optional[bool] = None
    """
    True, falls bei der Bildung des Durchschnittspreises für die Höchst- und Mindestpreisbetrachtung der Messpreis mit
    berücksichtigt wird
    """

    #: Typ des Messpreises
    messpreistyp: Optional[Messpreistyp] = None

    #: Im Preis bereits eingeschlossene Leistung (für Gas)
    kw_inklusive: Optional[Decimal] = None
    # todo: type decimal is most likely wrong: https://github.com/Hochfrequenz/BO4E-python/issues/327

    #: Intervall, indem die über "kwInklusive" hinaus abgenommene Leistung kostenpflichtig wird (z.B. je 5 kW 20 EURO)
    kw_weitere_mengen: Optional[Decimal] = None
    # todo: type decimal is most likely wrong: https://github.com/Hochfrequenz/BO4E-python/issues/327

    #: Höchstpreis für den Durchschnitts-Arbeitspreis NT
    hoechstpreis_n_t: Optional[Preis] = None
    #: Höchstpreis für den Durchschnitts-Arbeitspreis HT
    hoechstpreis_h_t: Optional[Preis] = None
    #: Mindestpreis für den Durchschnitts-Arbeitspreis
    mindestpreis: Optional[Preis] = None
    #: Liste mit zusätzlichen Preisen, beispielsweise Messpreise und/oder Leistungspreise
    zusatzpreise: Optional[List[Tarifpreis]] = None



import inspect
import re
from pathlib import Path
from typing import List, Optional, TypeVar

import pytest

from bo4e.enum import anrede
from bo4e.enum.strenum import StrEnum


class TestEnums:
    """
    A test class that checks the enum construction.
    """

    starts_with_whitespace_pattern = re.compile(r"^[\s\n]+")
    ends_with_whitespace_pattern = re.compile(r"[\s\n]+$")

    TEnum = TypeVar("TEnum", bound=StrEnum)

    @staticmethod
    def _get_all_enum_classes() -> List[TEnum]:
        """
        returns a list of all bo4e.enum classes
        """
        arbitrary_enum_module_path = Path(anrede.__file__)
        result: List[TestEnums.TEnum] = []
        for python_file in arbitrary_enum_module_path.parent.glob("*.py"):
            # don't ask me why. but it works.
            enum_module = __import__("bo4e.enum." + python_file.name.split(".")[0])
            for _, member in inspect.getmembers(enum_module.enum):
                if inspect.ismodule(member):
                    candidate = inspect.getmembers(member)[0][1]
                    if inspect.isclass(candidate):
                        result.append(candidate)  # type:ignore[arg-type]
        return result

    @staticmethod
    def _get_class_doc(enum_class: TEnum) -> Optional[str]:
        """
        asserts that enum class is a class and returns the class' docstring
        """
        assert inspect.isclass(enum_class)
        return inspect.getdoc(enum_class)

    def test_enum_classes_docstrings(self) -> None:
        """
        Tests that the docstrings of the enum classes do not start with whitespace or blank lines.
        """
        all_enums = TestEnums._get_all_enum_classes()  # type: ignore[var-annotated]
        assert len(all_enums) > 100  # just to be sure we're not using the wrong directory or path
        for enum_class in all_enums:
            docstring = TestEnums._get_class_doc(enum_class)
            assert docstring is not None
            assert not TestEnums.starts_with_whitespace_pattern.match(docstring)
            assert not TestEnums.ends_with_whitespace_pattern.match(docstring)



"""
Contains AufAbschlagProOrt class
and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from pydantic import conlist

from bo4e.com.aufabschlagstaffelproort import AufAbschlagstaffelProOrt
from bo4e.com.com import COM


class AufAbschlagProOrt(COM):
    """
    Mit dieser Komponente können Auf- und Abschläge verschiedener Typen im Zusammenhang
    mit örtlichen Gültigkeiten abgebildet werden.

    .. raw:: html

        <object data="../_static/images/bo4e/com/AufAbschlagProOrt.svg" type="image/svg+xml"></object>

    .. HINT::
        `AufAbschlagProOrt JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/AufAbschlagProOrt.json>`_

    """

    # required attributes
    #: Die Postleitzahl des Ortes für den der Aufschlag gilt.
    postleitzahl: str
    #: Der Ort für den der Aufschlag gilt.
    ort: str
    #: Die ene't-Netznummer des Netzes in dem der Aufschlag gilt.
    netznr: str
    #: Werte für die gestaffelten Auf/Abschläge mit regionaler Eingrenzung.
    staffeln: conlist(AufAbschlagstaffelProOrt, min_items=1)  # type: ignore[valid-type]



from typing import Any, Dict, List

import pytest
from pydantic import ValidationError

from bo4e.com.marktgebietinfo import MarktgebietInfo
from bo4e.com.standorteigenschaftengas import StandorteigenschaftenGas
from tests.serialization_helper import assert_serialization_roundtrip

example_standorteigenschaften_gas = StandorteigenschaftenGas(
    netzkontonummern=["GASPOOLNH700xxxx"],
    marktgebiete=[MarktgebietInfo(marktgebiet="Gaspool", marktgebietcode="37Z701133MH0000B")],
)


class TestStandorteigenschaftenGas:
    @pytest.mark.parametrize(
        "standorteigenschaftengas, expected_json_dict",
        [
            pytest.param(
                example_standorteigenschaften_gas,
                {
                    "netzkontonummern": ["GASPOOLNH700xxxx"],
                    "marktgebiete": [{"marktgebiet": "Gaspool", "marktgebietcode": "37Z701133MH0000B"}],
                },
            )
        ],
    )
    def test_standorteigenschaftengas_serialization_roundtrip(
        self, standorteigenschaftengas: StandorteigenschaftenGas, expected_json_dict: Dict[str, Any]
    ) -> None:
        assert_serialization_roundtrip(standorteigenschaftengas, expected_json_dict)

    def test_standorteigenschaftengas_missing_required_attributes(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = StandorteigenschaftenGas()  # type: ignore[call-arg]

        assert "2 validation errors" in str(excinfo.value)

    @pytest.mark.parametrize(
        "wrong_netzkontonummern, expected_error_message",
        [
            pytest.param(
                [],
                "ensure this value has at least 1 item",
            ),
            pytest.param(
                ["1", "2", "3"],
                "ensure this value has at most 2 items",
            ),
        ],
    )
    def test_standorteigenschaftengas_list_lenght_validation(
        self, wrong_netzkontonummern: List[str], expected_error_message: str
    ) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = StandorteigenschaftenGas(
                netzkontonummern=wrong_netzkontonummern,
                marktgebiete=[MarktgebietInfo(marktgebiet="Gaspool", marktgebietcode="37Z701133MH0000B")],
            )
        assert expected_error_message in str(excinfo.value)



"""
Contains PreisblattDienstleistung class and corresponding marshmallow schema for de-/serialization
"""
from typing import List, Optional

from bo4e.bo.preisblatt import Preisblatt
from bo4e.com.geraeteeigenschaften import Geraeteeigenschaften
from bo4e.enum.bilanzierungsmethode import Bilanzierungsmethode
from bo4e.enum.botyp import BoTyp
from bo4e.enum.dienstleistungstyp import Dienstleistungstyp

# pylint: disable=too-few-public-methods


class PreisblattDienstleistung(Preisblatt):
    """
    Variante des Preisblattmodells zur Abbildung der Preise für wahlfreie Dienstleistungen

    .. raw:: html

        <object data="../_static/images/bo4e/bo/PreisblattDienstleistung.svg" type="image/svg+xml"></object>

    .. HINT::
        `PreisblattDienstleistung JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/PreisblattDienstleistung.json>`_

    """

    bo_typ: BoTyp = BoTyp.PREISBLATTDIENSTLEISTUNG
    # required attributes (additional to those of Preisblatt)
    #: Die Preise gelten für Marktlokationen der angebebenen Bilanzierungsmethode
    bilanzierungsmethode: Bilanzierungsmethode
    #: Dienstleistung, für die der Preis abgebildet wird, z.B. Sperrung/Entsperrung
    basisdienstleistung: Dienstleistungstyp

    # optional attributes
    #: Hier kann der Preis auf bestimmte Geräte eingegrenzt werden. Z.B. auf die Zählergröße
    geraetedetails: Optional[Geraeteeigenschaften] = None

    #: Weitere Dienstleistungen, die im Preis enthalten sind
    inklusive_dienstleistungen: Optional[List[Dienstleistungstyp]] = None



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Kundengruppe(StrEnum):
    """
    Kundengruppe für eine Marktlokation (orientiert sich an den Standard-Lastprofilen).
    """

    RLM = "RLM"  #: Kunde mit registrierender Leistungsmessung (kein SLP)
    RLM_KOMMUNAL = "RLM_KOMMUNAL"  #: Kommunale Abnahmestelle leistungsgemessen
    SLP_KOMMUNAL = "SLP_KOMMUNAL"  #: Kommunale Abnahmestelle nicht leistungsgemessen
    SLP_S_G0 = "SLP_S_G0"  #: Gewerbe allgemein
    SLP_S_G1 = "SLP_S_G1"  #: Werktags
    SLP_S_G2 = "SLP_S_G2"  #: Verbrauch in Abendstunden
    SLP_S_G3 = "SLP_S_G3"  #: Gewerbe durchlaufend
    SLP_S_G4 = "SLP_S_G4"  #: Laden, Friseur
    SLP_S_G5 = "SLP_S_G5"  #: Bäckerei mit Backstube
    SLP_S_G6 = "SLP_S_G6"  #: Wochenendbetrieb
    SLP_S_G7 = "SLP_S_G7"  #: Mobilfunksendestation
    SLP_S_L0 = "SLP_S_L0"  #: Landwirtschaft allgemein
    SLP_S_L1 = "SLP_S_L1"  #: Landwirtschaft mit Milchwirtschaft/Nebenerwerbs-Tierzucht
    SLP_S_L2 = "SLP_S_L2"  #: Übrige Landwirtschaftsbetriebe
    SLP_S_H0 = "SLP_S_H0"  #: Haushalt allgemein
    SLP_S_SB = "SLP_S_SB"  #: Straßenbeleuchtung
    SLP_S_HZ = "SLP_S_HZ"  #: Nachtspeicherheizung
    SLP_S_WP = "SLP_S_WP"  #: Wärmepumpe
    SLP_S_EM = "SLP_S_EM"  #: Elektromobilität
    SLP_S_HZ_GEM = "SLP_S_HZ_GEM"  #: Nachtspeicherheizung gemeinsame Messung
    SLP_G_GKO = "SLP_G_GKO"  #: Gebietskörpersch., Kreditinst. u. Versich., Org. o. Erwerbszw. & öff. Einr.
    SLP_G_STANDARD = "SLP_G_STANDARD"  #: Standardkundengruppe für Gas
    SLP_G_GHA = "SLP_G_GHA"  #: Einzelhandel, Großhandel
    SLP_G_GMK = "SLP_G_GMK"  #: Metall, Kfz
    SLP_G_GBD = "SLP_G_GBD"  #: sonst. betr. Dienstleistungen
    SLP_G_GGA = "SLP_G_GGA"  #: Beherbergung
    SLP_G_GBH = "SLP_G_GBH"  #: Gaststätten
    SLP_G_GBA = "SLP_G_GBA"  #: Bäckereien
    SLP_G_GWA = "SLP_G_GWA"  #: Wäschereien
    SLP_G_GGB = "SLP_G_GGB"  #: Gartenbau
    SLP_G_GPD = "SLP_G_GPD"  #: Papier und Druck
    SLP_G_GMF = "SLP_G_GMF"  #: haushaltsähnliche Gewerbebetriebe
    SLP_G_HEF = "SLP_G_HEF"  #: Einfamilienhaushalt
    SLP_G_HMF = "SLP_G_HMF"  #: Mehrfamilienhaushalt
    SLP_G_HKO = "SLP_G_HKO"  #: Kochgas



"""
Contains Rufnummer class and corresponding marshmallow schema for de-/serialization
"""

from bo4e.com.com import COM
from bo4e.enum.rufnummernart import Rufnummernart

# pylint: disable=too-few-public-methods


class Rufnummer(COM):
    """
    Contains information to call or fax someone

    .. raw:: html

        <object data="../_static/images/bo4e/com/Rufnummer.svg" type="image/svg+xml"></object>

    .. HINT::
        `Rufnummer JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Rufnummer.json>`_

    """

    # required attributes
    #: Ausprägung der Nummer
    nummerntyp: Rufnummernart
    #: Die konkrete Nummer
    rufnummer: str



"""
Contains Steuerbetrag class
and corresponding marshmallow schema for de-/serialization
"""

from decimal import Decimal

from bo4e.com.com import COM
from bo4e.enum.steuerkennzeichen import Steuerkennzeichen
from bo4e.enum.waehrungscode import Waehrungscode

# pylint: disable=too-few-public-methods


class Steuerbetrag(COM):
    """
    Abbildung eines Steuerbetrages.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Steuerbetrag.svg" type="image/svg+xml"></object>

    .. HINT::
        `Steuerbetrag JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Steuerbetrag.json>`_

    """

    # required attributes
    #: Kennzeichnung des Steuersatzes, bzw. Verfahrens.
    steuerkennzeichen: Steuerkennzeichen
    #: Nettobetrag für den die Steuer berechnet wurde. Z.B. 100
    basiswert: Decimal
    #: Aus dem Basiswert berechnete Steuer. Z.B. 19 (bei UST_19)
    steuerwert: Decimal
    #: Währung. Z.B. Euro.
    waehrung: Waehrungscode



from decimal import Decimal
from typing import Any, Dict

import pytest

from bo4e.com.geraet import Geraet
from bo4e.com.geraeteeigenschaften import Geraeteeigenschaften
from bo4e.com.menge import Menge
from bo4e.com.tarifeinschraenkung import Tarifeinschraenkung
from bo4e.enum.geraetemerkmal import Geraetemerkmal
from bo4e.enum.geraetetyp import Geraetetyp
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.voraussetzungen import Voraussetzungen
from tests.serialization_helper import assert_serialization_roundtrip

example_tarifeinschraenkung = Tarifeinschraenkung(
    zusatzprodukte=["foo", "bar"],
    voraussetzungen=[Voraussetzungen.ALTVERTRAG, Voraussetzungen.DIREKTVERTRIEB],
    einschraenkungzaehler=[
        Geraet(
            geraetenummer="0815",
            geraeteeigenschaften=Geraeteeigenschaften(
                geraetemerkmal=Geraetemerkmal.GAS_G1000,
                geraetetyp=Geraetetyp.MULTIPLEXANLAGE,
            ),
        ),
        Geraet(geraetenummer="197foo"),
    ],
    einschraenkungleistung=[
        Menge(wert=Decimal(12.5), einheit=Mengeneinheit.MWH),
        Menge(wert=Decimal(30), einheit=Mengeneinheit.KWH),
    ],
)


class TestTarifeinschraenkung:
    @pytest.mark.parametrize(
        "tarifeinschraenkung, expected_json_dict",
        [
            pytest.param(
                Tarifeinschraenkung(),
                {
                    "zusatzprodukte": None,
                    "voraussetzungen": None,
                    "einschraenkungzaehler": None,
                    "einschraenkungleistung": None,
                },
                id="minimal attributes",
            ),
            pytest.param(
                Tarifeinschraenkung(
                    zusatzprodukte=["foo", "bar"],
                    voraussetzungen=[Voraussetzungen.ALTVERTRAG, Voraussetzungen.DIREKTVERTRIEB],
                    einschraenkungzaehler=[
                        Geraet(
                            geraetenummer="0815",
                            geraeteeigenschaften=Geraeteeigenschaften(
                                geraetemerkmal=Geraetemerkmal.GAS_G1000,
                                geraetetyp=Geraetetyp.MULTIPLEXANLAGE,
                            ),
                        ),
                        Geraet(geraetenummer="197foo"),
                    ],
                    einschraenkungleistung=[
                        Menge(wert=Decimal(12.5), einheit=Mengeneinheit.MWH),
                        Menge(wert=Decimal(30), einheit=Mengeneinheit.KWH),
                    ],
                ),
                {
                    "zusatzprodukte": ["foo", "bar"],
                    "voraussetzungen": ["ALTVERTRAG", "DIREKTVERTRIEB"],
                    "einschraenkungzaehler": [
                        {
                            "geraetenummer": "0815",
                            "geraeteeigenschaften": {
                                "geraetemerkmal": "GAS_G1000",
                                "geraetetyp": Geraetetyp.MULTIPLEXANLAGE,
                            },
                        },
                        {
                            "geraetenummer": "197foo",
                            "geraeteeigenschaften": None,
                        },
                    ],
                    "einschraenkungleistung": [
                        {
                            "wert": Decimal("12.5"),
                            "einheit": Mengeneinheit.MWH,
                        },
                        {
                            "wert": Decimal("30"),
                            "einheit": Mengeneinheit.KWH,
                        },
                    ],
                },
                id="maximal attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(
        self, tarifeinschraenkung: Tarifeinschraenkung, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of Tarifeinschraenkung
        """
        assert_serialization_roundtrip(tarifeinschraenkung, expected_json_dict)



import pytest
from pydantic import ValidationError

from bo4e.com.tarifpreispositionproort import TarifpreispositionProOrt
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_tarifpreisstaffelproort import example_tarifpreisstaffelproort

example_tarifpreispositionproort = TarifpreispositionProOrt(
    postleitzahl="82031",
    ort="Grünwald",
    netznr="0815",
    preisstaffeln=[example_tarifpreisstaffelproort],
)


class TestTarifpreispositionProOrt:
    @pytest.mark.parametrize(
        "tarifpreispositionproort",
        [
            pytest.param(
                example_tarifpreispositionproort,
                id="minimal and maximal attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, tarifpreispositionproort: TarifpreispositionProOrt) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(tarifpreispositionproort)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = TarifpreispositionProOrt()  # type: ignore[call-arg]

        assert "4 validation errors" in str(excinfo.value)



"""
Contains Ausschreibung class and corresponding marshmallow schema for de-/serialization
"""
from datetime import datetime

# pylint: disable=too-few-public-methods, too-many-instance-attributes
# pylint: disable=no-name-in-module
from typing import Optional

from pydantic import conlist

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.com.ausschreibungslos import Ausschreibungslos
from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.ausschreibungsportal import Ausschreibungsportal
from bo4e.enum.ausschreibungsstatus import Ausschreibungsstatus
from bo4e.enum.ausschreibungstyp import Ausschreibungstyp
from bo4e.enum.botyp import BoTyp


class Ausschreibung(Geschaeftsobjekt):
    """
    Das BO Ausschreibung dient zur detaillierten Darstellung von ausgeschriebenen Energiemengen in der Energiewirtschaft

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Ausschreibung.svg" type="image/svg+xml"></object>

    .. HINT::
        `Ausschreibung JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Ausschreibung.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.AUSSCHREIBUNG
    #: Vom Herausgeber der Ausschreibung vergebene eindeutige Nummer
    ausschreibungsnummer: str
    #: Aufzählung für die Typisierung von Ausschreibungen
    ausschreibungstyp: Ausschreibungstyp
    #: Bezeichnungen für die Ausschreibungsphasen
    ausschreibungsstatus: Ausschreibungsstatus
    #: Kennzeichen, ob die Ausschreibung kostenpflichtig ist
    kostenpflichtig: bool
    #: Gibt den Veröffentlichungszeitpunkt der Ausschreibung an
    veroeffentlichungszeitpunkt: datetime
    ausschreibender: Geschaeftspartner
    """
    Mit diesem Objekt können Geschäftspartner übertragen werden.
    Sowohl Unternehmen, als auch Privatpersonen können Geschäftspartner sein
    """
    abgabefrist: Zeitraum
    """
    Diese Komponente wird zur Abbildung von Zeiträumen in Form von Dauern oder der Angabe von Start und Ende verwendet.
    Es muss daher entweder eine Dauer oder ein Zeitraum in Form von Start und Ende angegeben sein
    """
    bindefrist: Zeitraum
    """
    Diese Komponente wird zur Abbildung von Zeiträumen in Form von Dauern oder der Angabe von Start und Ende verwendet.
    Es muss daher entweder eine Dauer oder ein Zeitraum in Form von Start und Ende angegeben sein
    """
    #: Die einzelnen Lose, aus denen sich die Ausschreibung zusammensetzt
    lose: conlist(Ausschreibungslos, min_items=1)  # type: ignore[valid-type]

    # optional attributes
    #: Aufzählung der unterstützten Ausschreibungsportale
    ausschreibungportal: Optional[Ausschreibungsportal] = None
    #: Internetseite, auf der die Ausschreibung veröffentlicht wurde (falls vorhanden)
    webseite: Optional[str] = None



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Messwertstatus(StrEnum):
    """
    Der Status eines Zählerstandes
    """

    ABGELESEN = "ABGELESEN"  #: ABGELESEN
    ERSATZWERT = "ERSATZWERT"  #: ERSATZWERT
    VORSCHLAGSWERT = "VORSCHLAGSWERT"  #: VORSCHLAGSWERT
    NICHT_VERWENDBAR = "NICHT_VERWENDBAR"  #: NICHT_VERWENDBAR
    PROGNOSEWERT = "PROGNOSEWERT"  #: PROGNOSEWERT
    VORLAEUFIGERWERT = "VORLAEUFIGERWERT"  #: VORLAEUFIGERWERT
    ENERGIEMENGESUMMIERT = "ENERGIEMENGESUMMIERT"  #: ENERGIEMENGESUMMIERT
    FEHLT = "FEHLT"  #: FEHLT



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Zaehlertyp(StrEnum):
    """
    Bei diesem Enum handelt es sich um die Abbildung von Zählertypen der Sparten Strom und Gas.
    """

    DREHSTROMZAEHLER = "DREHSTROMZAEHLER"  #: Drehstromzähler
    BALGENGASZAEHLER = "BALGENGASZAEHLER"  #: Balgengaszähler
    DREHKOLBENZAEHLER = "DREHKOLBENZAEHLER"  #: Drehkolbengaszähler
    SMARTMETER = "SMARTMETER"  #: Smart Meter Zähler
    LEISTUNGSZAEHLER = "LEISTUNGSZAEHLER"  #: leistungsmessender Zähler
    MAXIMUMZAEHLER = "MAXIMUMZAEHLER"  #: Maximumzähler
    TURBINENRADGASZAEHLER = "TURBINENRADGASZAEHLER"  #: Turbinenradgaszähler
    ULTRASCHALLGASZAEHLER = "ULTRASCHALLGASZAEHLER"  #: Ultraschallgaszähler
    WECHSELSTROMZAEHLER = "WECHSELSTROMZAEHLER"  #: Wechselstromzähler



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Sparte(StrEnum):
    """
    Unterscheidungsmöglichkeiten für die Sparte.
    """

    STROM = "STROM"
    GAS = "GAS"
    FERNWAERME = "FERNWAERME"
    NAHWAERME = "NAHWAERME"
    WASSER = "WASSER"
    ABWASSER = "ABWASSER"
    STROM_UND_GAS = "STROM_UND_GAS"



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Geraetemerkmal(StrEnum):
    """
    Auflistung möglicher abzurechnender Gerätetypen.
    """

    EINTARIF = "EINTARIF"  #: Eintarifzähler
    ZWEITARIF = "ZWEITARIF"  #: Zweitarifzähler
    MEHRTARIF = "MEHRTARIF"  #: Mehrtarifzähler
    GAS_G2_5 = "GAS_G2_5"  #: Gaszähler Größe G2.5
    GAS_G4 = "GAS_G4"  #: Gaszähler Größe G4
    GAS_G6 = "GAS_G6"  #: Gaszähler Größe G6
    GAS_G10 = "GAS_G10"  #: Gaszähler Größe G10
    GAS_G16 = "GAS_G16"  #: Gaszähler Größe G16
    GAS_G25 = "GAS_G25"  #: Gaszähler Größe G25
    GAS_G40 = "GAS_G40"  #: Gaszähler Größe G40
    GAS_G65 = "GAS_G65"  #: Gaszähler Größe G65
    GAS_G100 = "GAS_G100"  #: Gaszähler Größe G100
    GAS_G160 = "GAS_G160"  #: Gaszähler Größe G160
    GAS_G250 = "GAS_G250"  #: Gaszähler Größe G250
    GAS_G400 = "GAS_G400"  #: Gaszähler Größe G400
    GAS_G650 = "GAS_G650"  #: Gaszähler Größe G650
    GAS_G1000 = "GAS_G1000"  #: Gaszähler Größe G1000
    GAS_G1600 = "GAS_G1600"  #: Gaszähler Größe G1600
    GAS_G2500 = "GAS_G2500"  #: Gaszähler Größe G2500
    IMPULSGEBER_G4_G100 = "IMPULSGEBER_G4_G100"  #: Impulsgeber für Zähler G4 - G100
    IMPULSGEBER_G100 = "IMPULSGEBER_G100"  #: Impulsgeber für Zähler größer G100
    TELEFONANSCHLUSS = "TELEFONANSCHLUSS"  #: Telefonanschluss
    MODEM_GSM = "MODEM_GSM"  #: Modem-GSM
    MODEM_GPRS = "MODEM_GPRS"  #: Modem-GPRS
    MODEM_FUNK = "MODEM_FUNK"  #: Modem-Funk
    MODEM_GSM_O_LG = "MODEM_GSM_O_LG"  #: vom Messstellenbetreiber beigestelltes GSM-Modem ohne Lastgangmessung
    MODEM_GSM_M_LG = "MODEM_GSM_M_LG"  #: vom Messstellenbetreiber beigestelltes GSM-Modem mit Lastgangmessung
    MODEM_FESTNETZ = "MODEM_FESTNETZ"  #: vom Messstellenbetreiber beigestelltes Festnetz-Modem
    MODEM_GPRS_M_LG = "MODEM_GPRS_M_LG"  #: vom Messstellenbetreiber bereitgestelltes GPRS-Modem Lastgangmessung
    PLC_KOM = "PLC_KOM"  #: PLC-Kom.-Einrichtung (Powerline Communication)
    ETHERNET_KOM = "ETHERNET_KOM"  #: Ethernet-Kom.-Einricht. LAN/WLAN
    DSL_KOM = "DSL_KOM"  #: DSL-Kommunikationseinrichtung
    LTE_KOM = "LTE_KOM"  #: LTE-Kommunikationseinrichtung
    KOMPAKT_MU = "KOMPAKT_MU"  #: Kompaktmengenumwerter
    SYSTEM_MU = "SYSTEM_MU"  #: Systemmengenumwerter
    TEMPERATUR_MU = "TEMPERATUR_MU"  #: Temperaturmengenumwerter
    ZUSTANDS_MU = "ZUSTANDS_MU"  #: Zustandsmengenumwerter
    UNBESTIMMT = "UNBESTIMMT"  #: Unbestimmtes Merkmal



from datetime import datetime, timezone
from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.betrag import Betrag
from bo4e.com.fremdkostenposition import Fremdkostenposition
from bo4e.com.preis import Preis
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.preisstatus import Preisstatus
from bo4e.enum.waehrungscode import Waehrungscode
from bo4e.enum.waehrungseinheit import Waehrungseinheit
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_menge import example_menge
from tests.test_sigmoidparameter import example_sigmoidparameter


class TestFremdkostenposition:
    @pytest.mark.parametrize(
        "fremdkostenposition, expected_json_dict",
        [
            pytest.param(
                Fremdkostenposition(
                    positionstitel="Mudders Preisstaffel",
                    artikelbezeichnung="Dei Mudder ihr Preisstaffel",
                    einzelpreis=Preis(
                        wert=Decimal(3.50),
                        einheit=Waehrungseinheit.EUR,
                        bezugswert=Mengeneinheit.KWH,
                        status=Preisstatus.ENDGUELTIG,
                    ),
                    betrag_kostenposition=Betrag(
                        waehrung=Waehrungscode.EUR,
                        wert=Decimal(12.5),
                    ),
                ),
                {
                    "marktpartnercode": None,
                    "positionstitel": "Mudders Preisstaffel",
                    "einzelpreis": {
                        "wert": Decimal("3.5"),
                        "einheit": Waehrungseinheit.EUR,
                        "bezugswert": Mengeneinheit.KWH,
                        "status": Preisstatus.ENDGUELTIG,
                    },
                    "bis": None,
                    "menge": None,
                    "zeitmenge": None,
                    "artikelbezeichnung": "Dei Mudder ihr Preisstaffel",
                    "marktpartnername": None,
                    "artikeldetail": None,
                    "von": None,
                    "linkPreisblatt": None,
                    "betragKostenposition": {"wert": Decimal("12.5"), "waehrung": Waehrungseinheit.EUR},
                    "gebietcodeEic": None,
                },
                id="only required attributes",
            ),
            pytest.param(
                Fremdkostenposition(
                    positionstitel="Vadders Preisstaffel",
                    von=datetime(2013, 5, 1, tzinfo=timezone.utc),
                    bis=datetime(2014, 5, 1, tzinfo=timezone.utc),
                    artikelbezeichnung="Deim Vadder sei Preisstaffel",
                    link_preisblatt="http://foo.bar/",
                    zeitmenge=example_menge,
                    einzelpreis=Preis(
                        wert=Decimal(3.50),
                        einheit=Waehrungseinheit.EUR,
                        bezugswert=Mengeneinheit.KWH,
                        status=Preisstatus.ENDGUELTIG,
                    ),
                    betrag_kostenposition=Betrag(
                        waehrung=Waehrungscode.EUR,
                        wert=Decimal(12.5),
                    ),
                    menge=example_menge,
                    artikeldetail="foo",
                    marktpartnercode="986543210123",
                    marktpartnername="Mein MP",
                    gebietcode_eic="not an eic code but validation will follow in ticket 146",
                ),
                {
                    "artikelbezeichnung": "Deim Vadder sei Preisstaffel",
                    "artikeldetail": "foo",
                    "marktpartnername": "Mein MP",
                    "einzelpreis": {
                        "bezugswert": Mengeneinheit.KWH,
                        "status": Preisstatus.ENDGUELTIG,
                        "wert": Decimal("3.5"),
                        "einheit": Waehrungseinheit.EUR,
                    },
                    "menge": {
                        "wert": Decimal("3.410000000000000142108547152020037174224853515625"),
                        "einheit": Mengeneinheit.MWH,
                    },
                    "zeitmenge": {
                        "wert": Decimal("3.410000000000000142108547152020037174224853515625"),
                        "einheit": Mengeneinheit.MWH,
                    },
                    "marktpartnercode": "986543210123",
                    "bis": datetime(2014, 5, 1, 0, 0, tzinfo=timezone.utc),
                    "positionstitel": "Vadders Preisstaffel",
                    "von": datetime(2013, 5, 1, 0, 0, tzinfo=timezone.utc),
                    "betragKostenposition": {"wert": Decimal("12.5"), "waehrung": Waehrungseinheit.EUR},
                    "gebietcodeEic": "not an eic code but validation will follow in ticket 146",
                    "linkPreisblatt": "http://foo.bar/",
                },
                id="required and optional attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(
        self, fremdkostenposition: Fremdkostenposition, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of Fremdkostenposition.
        """
        assert_serialization_roundtrip(fremdkostenposition, expected_json_dict)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Fremdkostenposition()  # type: ignore[call-arg]

        assert "4 validation errors" in str(excinfo.value)



"""
Contains AufAbschlagRegional and corresponding marshmallow schema for de-/serialization
"""

from typing import List, Optional

# pylint: disable=too-few-public-methods, too-many-instance-attributes
# pylint: disable=no-name-in-module
from pydantic import conlist

from bo4e.com.aufabschlagproort import AufAbschlagProOrt
from bo4e.com.com import COM
from bo4e.com.energiemix import Energiemix
from bo4e.com.preisgarantie import Preisgarantie
from bo4e.com.tarifeinschraenkung import Tarifeinschraenkung
from bo4e.com.vertragskonditionen import Vertragskonditionen
from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.aufabschlagstyp import AufAbschlagstyp
from bo4e.enum.aufabschlagsziel import AufAbschlagsziel
from bo4e.enum.waehrungseinheit import Waehrungseinheit


# pylint: disable=R0801
class AufAbschlagRegional(COM):
    """
    Mit dieser Komponente können Auf- und Abschläge verschiedener Typen
    im Zusammenhang mit regionalen Gültigkeiten abgebildet werden.
    Hier sind auch die Auswirkungen auf verschiedene Tarifparameter modelliert,
    die sich durch die Auswahl eines Auf- oder Abschlags ergeben.

    .. raw:: html

        <object data="../_static/images/bo4e/com/AufAbschlagRegional.svg" type="image/svg+xml"></object>

    .. HINT::
        `AufAbschlagRegional JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/AufAbschlagRegional.json>`_

    """

    # required attributess
    #: Bezeichnung des Auf-/Abschlags
    bezeichnung: str
    #: Werte für die gestaffelten Auf/Abschläge mit regionaler Eingrenzung
    betraege: conlist(AufAbschlagProOrt, min_items=1)  # type: ignore[valid-type]

    # optional attributes
    #: Beschreibung zum Auf-/Abschlag
    beschreibung: Optional[str] = None
    #:Typ des Aufabschlages (z.B. absolut oder prozentual)
    auf_abschlagstyp: Optional[AufAbschlagstyp] = None
    #: Diesem Preis oder den Kosten ist der Auf/Abschlag zugeordnet. Z.B. Arbeitspreis, Gesamtpreis etc.
    auf_abschlagsziel: Optional[AufAbschlagsziel] = None
    #: Gibt an in welcher Währungseinheit der Auf/Abschlag berechnet wird. Euro oder Ct.
    einheit: Optional[Waehrungseinheit] = None
    #: Internetseite, auf der die Informationen zum Auf-/Abschlag veröffentlicht sind
    website: Optional[str] = None
    #: Zusatzprodukte, die nur in Kombination mit diesem AufAbschlag erhältlich sind
    zusatzprodukte: Optional[List[str]] = None
    #: Voraussetzungen, die erfüllt sein müssen, damit dieser AufAbschlag zur Anwendung kommen kann
    voraussetzungen: Optional[List[str]] = None
    #: Durch die Anwendung des Auf/Abschlags kann eine Änderung des Tarifnamens auftreten.
    tarifnamensaenderungen: Optional[str] = None
    #: Zeitraum, in dem der Abschlag zur Anwendung kommen kann
    gueltigkeitszeitraum: Optional[Zeitraum] = None
    energiemixaenderung: Optional[Energiemix] = None
    """ Der Energiemix kann sich durch einen AufAbschlag ändern (z.B. zwei Cent Aufschlag für Ökostrom:
    Sollte dies der Fall sein, wird hier die neue Zusammensetzung des Energiemix angegeben."""
    vertagskonditionsaenderung: Optional[Vertragskonditionen] = None
    """ Änderungen in den Vertragskonditionen. Falls in dieser Komponenten angegeben,
    werden die Tarifparameter hiermit überschrieben."""
    garantieaenderung: Optional[Preisgarantie] = None
    """ Änderungen in den Garantievereinbarungen. Falls in dieser Komponenten angegeben,
    werden die Tarifparameter hiermit überschrieben."""
    einschraenkungsaenderung: Optional[Tarifeinschraenkung] = None
    """ Änderungen in den Einschränkungen zum Tarif. Falls in dieser Komponenten angegeben,
    werden die Tarifparameter hiermit überschrieben."""



from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.bo.lastgang import LastgangKompakt
from bo4e.com.zeitintervall import Zeitintervall
from bo4e.enum.lokationstyp import Lokationstyp
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.sparte import Sparte
from bo4e.enum.zeiteinheit import Zeiteinheit
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_tagesvektor import example_tagesvektor, example_tagesvektor_json


class TestLastgangKompakt:
    @pytest.mark.parametrize(
        "lastgang_kompakt, expected_json_dict",
        [
            pytest.param(
                LastgangKompakt(
                    version="1.1",
                    sparte=Sparte.STROM,
                    lokations_id="DE0000011111222223333344444555556",
                    obis_kennzahl="1-0:1.8.1",
                    lokationstyp=Lokationstyp.MELO,
                    messgroesse=Mengeneinheit.KWH,
                    zeitintervall=Zeitintervall(
                        wert=1,
                        zeiteinheit=Zeiteinheit.VIERTEL_STUNDE,
                    ),
                    tagesvektoren=[example_tagesvektor],
                ),
                {
                    "version": "1.1",
                    "sparte": Sparte.STROM,
                    "lokationstyp": Lokationstyp.MELO,
                    "messgroesse": Mengeneinheit.KWH,
                    "zeitintervall": {"zeiteinheit": "VIERTEL_STUNDE", "wert": 1},
                    "tagesvektoren": [example_tagesvektor_json],
                    "versionstruktur": "2",
                    "externeReferenzen": [],
                    "lokationsId": "DE0000011111222223333344444555556",
                    "boTyp": "LASTGANG_KOMPAKT",
                    "obisKennzahl": "1-0:1.8.1",
                },
            ),
        ],
    )
    def test_serialization_roundtrip(
        self, lastgang_kompakt: LastgangKompakt, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of LastgangKompakt.
        """
        assert_serialization_roundtrip(lastgang_kompakt, expected_json_dict)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = LastgangKompakt()  # type: ignore[call-arg]

        assert "6 validation errors" in str(excinfo.value)



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class NNRechnungsart(StrEnum):
    """
    Abbildung verschiedener in der INVOIC angegebenen Rechnungsarten.
    """

    HANDELSRECHNUNG = "HANDELSRECHNUNG"  #: HANDELSRECHNUNG
    SELBSTAUSGESTELLT = "SELBSTAUSGESTELLT"  #: SELBSTAUSGESTELLT



from datetime import datetime, timezone
from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.bo.marktlokation import Marktlokation
from bo4e.com.adresse import Adresse
from bo4e.com.angebotsposition import Angebotsposition
from bo4e.com.angebotsteil import Angebotsteil
from bo4e.com.betrag import Betrag
from bo4e.com.menge import Menge
from bo4e.com.preis import Preis
from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.bilanzierungsmethode import Bilanzierungsmethode
from bo4e.enum.botyp import BoTyp
from bo4e.enum.energierichtung import Energierichtung
from bo4e.enum.landescode import Landescode
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.netzebene import Netzebene
from bo4e.enum.sparte import Sparte
from bo4e.enum.waehrungscode import Waehrungscode
from bo4e.enum.waehrungseinheit import Waehrungseinheit
from tests.serialization_helper import assert_serialization_roundtrip

example_angebotsteil: Angebotsteil = Angebotsteil(
    positionen=[
        Angebotsposition(
            positionsbezeichnung="teststring",
            positionsmenge=Menge(wert=Decimal(4000), einheit=Mengeneinheit.KWH),
            positionspreis=Preis(wert=Decimal(0.2456), einheit=Waehrungseinheit.EUR, bezugswert=Mengeneinheit.KWH),
            positionskosten=Betrag(
                waehrung=Waehrungscode.EUR,
                wert=Decimal(98240),
            ),
        )
    ],
)

example_angebotsteil_json = {
    "positionen": [
        {
            "positionsbezeichnung": "teststring",
            "positionsmenge": {"wert": Decimal("4000"), "einheit": Mengeneinheit.KWH},
            "positionskosten": {"waehrung": Waehrungseinheit.EUR, "wert": Decimal("98240")},
            "positionspreis": {
                "bezugswert": Mengeneinheit.KWH,
                "status": None,
                "wert": Decimal("0.2456000000000000127453603226967970840632915496826171875"),
                "einheit": Waehrungseinheit.EUR,
            },
        },
    ],
    "anfrageSubreferenz": None,
    "lieferstellenangebotsteil": None,
    "gesamtmengeangebotsteil": None,
    "gesamtkostenangebotsteil": None,
    "lieferzeitraum": None,
}


class TestAngebotsteil:
    @pytest.mark.parametrize(
        "angebotsteil, expected_json_dict",
        [
            pytest.param(
                Angebotsteil(
                    positionen=[
                        Angebotsposition(
                            positionsbezeichnung="testtring",
                            positionsmenge=Menge(wert=Decimal(4000), einheit=Mengeneinheit.KWH),
                            positionspreis=Preis(
                                wert=Decimal(0.2456), einheit=Waehrungseinheit.EUR, bezugswert=Mengeneinheit.KWH
                            ),
                            positionskosten=Betrag(
                                waehrung=Waehrungscode.EUR,
                                wert=Decimal(98240),
                            ),
                        )
                    ],
                    anfrage_subreferenz="teststring",
                    lieferstellenangebotsteil=[
                        Marktlokation(
                            marktlokations_id="51238696781",
                            sparte=Sparte.GAS,
                            lokationsadresse=Adresse(
                                postleitzahl="82031",
                                ort="Grünwald",
                                hausnummer="27A",
                                strasse="Nördliche Münchner Straße",
                            ),
                            energierichtung=Energierichtung.EINSP,
                            bilanzierungsmethode=Bilanzierungsmethode.PAUSCHAL,
                            unterbrechbar=True,
                            netzebene=Netzebene.NSP,
                        )
                    ],
                    gesamtmengeangebotsteil=Menge(wert=Decimal(4000), einheit=Mengeneinheit.KWH),
                    gesamtkostenangebotsteil=Betrag(
                        waehrung=Waehrungscode.EUR,
                        wert=Decimal(98240),
                    ),
                    lieferzeitraum=Zeitraum(
                        startdatum=datetime(2020, 1, 1, tzinfo=timezone.utc),
                        enddatum=datetime(2020, 4, 1, tzinfo=timezone.utc),
                    ),
                ),
                {
                    "positionen": [
                        {
                            "positionsbezeichnung": "testtring",
                            "positionsmenge": {"wert": Decimal("4000"), "einheit": Mengeneinheit.KWH},
                            "positionskosten": {"waehrung": Waehrungseinheit.EUR, "wert": Decimal("98240")},
                            "positionspreis": {
                                "bezugswert": Mengeneinheit.KWH,
                                "status": None,
                                "wert": Decimal("0.2456000000000000127453603226967970840632915496826171875"),
                                "einheit": Waehrungseinheit.EUR,
                            },
                        },
                    ],
                    "lieferstellenangebotsteil": [
                        {
                            "marktlokationsId": "51238696781",
                            "sparte": Sparte.GAS,
                            "lokationsadresse": {
                                "postleitzahl": "82031",
                                "ort": "Grünwald",
                                "hausnummer": "27A",
                                "strasse": "Nördliche Münchner Straße",
                                "adresszusatz": None,
                                "postfach": None,
                                "coErgaenzung": None,
                                "landescode": Landescode.DE,  # type: ignore[attr-defined]
                            },
                            "energierichtung": Energierichtung.EINSP,
                            "bilanzierungsmethode": Bilanzierungsmethode.PAUSCHAL,
                            "unterbrechbar": True,
                            "netzebene": Netzebene.NSP,
                            "netzgebietsnr": None,
                            "versionstruktur": "2",
                            "katasterinformation": None,
                            "bilanzierungsgebiet": None,
                            "grundversorgercodenr": None,
                            "endkunde": None,
                            "geoadresse": None,
                            "verbrauchsart": None,
                            "netzbetreibercodenr": None,
                            "gebietstyp": None,
                            "gasqualitaet": None,
                            "zugehoerigeMesslokation": None,
                            "kundengruppen": None,
                            "externeReferenzen": [],
                            "boTyp": BoTyp.MARKTLOKATION,
                        }
                    ],
                    "gesamtmengeangebotsteil": {"wert": Decimal("4000"), "einheit": Mengeneinheit.KWH},
                    "gesamtkostenangebotsteil": {"waehrung": Waehrungseinheit.EUR, "wert": Decimal("98240")},
                    "anfrageSubreferenz": "teststring",
                    "lieferzeitraum": {
                        "startdatum": datetime(2020, 1, 1, tzinfo=timezone.utc),
                        "endzeitpunkt": None,
                        "einheit": None,
                        "enddatum": datetime(2020, 4, 1, tzinfo=timezone.utc),
                        "startzeitpunkt": None,
                        "dauer": None,
                    },
                },
                id="maximal attributes",
            ),
            pytest.param(
                example_angebotsteil,
                example_angebotsteil_json,
                id="minimal attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, angebotsteil: Angebotsteil, expected_json_dict: Dict[str, Any]) -> None:
        """
        Test de-/serialisation of Angebotsteil with minimal attributes.
        """
        assert_serialization_roundtrip(angebotsteil, expected_json_dict)

    def test_angebotsteil_positionen_required(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Angebotsteil(positionen=[])

        assert "1 validation error" in str(excinfo.value)
        assert "ensure this value has at least 1 item" in str(excinfo.value)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Angebotsteil()  # type: ignore[call-arg]
        assert "1 validation error" in str(excinfo.value)



"""
Contains Ansprechpartner class
and corresponding marshmallow schema for de-/serialization
"""
from typing import Optional

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.com.adresse import Adresse
from bo4e.com.rufnummer import Rufnummer
from bo4e.com.zustaendigkeit import Zustaendigkeit
from bo4e.enum.anrede import Anrede
from bo4e.enum.botyp import BoTyp
from bo4e.enum.titel import Titel

# pylint: disable=too-many-instance-attributes, too-few-public-methods


class Ansprechpartner(Geschaeftsobjekt):
    """
    Object containing information about a Ansprechpartner

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Ansprechpartner.svg" type="image/svg+xml"></object>

    .. HINT::
        `Ansprechpartner JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Ansprechpartner.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.ANSPRECHPARTNER
    nachname: str  #: Nachname (Familienname) des Ansprechpartners
    geschaeftspartner: Geschaeftspartner  #: Der Geschäftspartner, für den dieser Ansprechpartner modelliert wird

    # optional attributes
    anrede: Optional[Anrede] = None  #: Mögliche Anrede des Ansprechpartners
    individuelle_anrede: Optional[str] = None
    """
    Im Falle einer nicht standardisierten Anrede kann hier eine frei definierbare Anrede vorgegeben werden.
    Beispiel: "Sehr geehrte Frau Müller, sehr geehrter Herr Dr. Müller"
    """

    titel: Optional[Titel] = None  #: Möglicher Titel des Ansprechpartners
    vorname: Optional[str] = None  #: Vorname des Ansprechpartners
    e_mail_adresse: Optional[str] = None  #: E-Mail Adresse
    kommentar: Optional[str] = None  #: Weitere Informationen zum Ansprechpartner
    #: Adresse des Ansprechpartners, falls diese von der Adresse des Geschäftspartners abweicht
    adresse: Optional[Adresse] = None
    #: Liste der Telefonnummern, unter denen der Ansprechpartner erreichbar ist
    rufnummer: Optional[Rufnummer] = None  # todo: make this a list and rename to rufnummern
    #: Liste der Abteilungen und Zuständigkeiten des Ansprechpartners
    zustaendigkeit: Optional[Zustaendigkeit] = None  # todo: make this a list and rename to "zustaendigkeiten"



from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.regionskriterium import Regionskriterium
from bo4e.enum.gueltigkeitstyp import Gueltigkeitstyp
from bo4e.enum.regionskriteriumtyp import Regionskriteriumtyp
from tests.serialization_helper import assert_serialization_roundtrip

example_regionskriterium = Regionskriterium(
    regionskriteriumtyp=Regionskriteriumtyp.REGELGEBIET_NAME,
    gueltigkeitstyp=Gueltigkeitstyp.NICHT_IN,
    wert="Was ist ein Regionskriterium?",
)


class TestRegionskriterium:
    @pytest.mark.parametrize(
        "regionskriterium, expected_json_dict",
        [
            pytest.param(
                example_regionskriterium,
                {
                    "gueltigkeitstyp": "NICHT_IN",
                    "regionskriteriumtyp": "REGELGEBIET_NAME",
                    "wert": "Was ist ein Regionskriterium?",
                },
            ),
        ],
    )
    def test_regionskriterium_serialization_roundtrip(
        self, regionskriterium: Regionskriterium, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of Regionskriterium with minimal attributes.
        """
        assert_serialization_roundtrip(regionskriterium, expected_json_dict)

    def test_regionskriterium_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Regionskriterium()  # type: ignore[call-arg]
        assert "3 validation errors" in str(excinfo.value)



"""
Contains Katasteradresse class
and corresponding marshmallow schema for de-/serialization
"""

from bo4e.com.com import COM

# pylint: disable=too-few-public-methods


class Katasteradresse(COM):
    """
    Dient der Adressierung über die Liegenschafts-Information.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Katasteradresse.svg" type="image/svg+xml"></object>

    .. HINT::
        `Katasteradresse JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Katasteradresse.json>`_

    """

    gemarkung_flur: str
    flurstueck: str



"""
Contains Tarifeinschraenkung and corresponding marshmallow schema for de-/serialization
"""
from typing import List, Optional

from bo4e.com.com import COM
from bo4e.com.geraet import Geraet
from bo4e.com.menge import Menge
from bo4e.enum.voraussetzungen import Voraussetzungen

# pylint: disable=too-few-public-methods


class Tarifeinschraenkung(COM):
    """
    Mit dieser Komponente werden Einschränkungen für die Anwendung von Tarifen modelliert.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Tarifeinschraenkung.svg" type="image/svg+xml"></object>

    .. HINT::
        `Tarifeinschraenkung JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Tarifeinschraenkung.json>`_

    """

    # optional attributes
    #: Weitere Produkte, die gemeinsam mit diesem Tarif bestellt werden können
    zusatzprodukte: Optional[List[str]] = None
    #: Voraussetzungen, die erfüllt sein müssen, damit dieser Tarif zur Anwendung kommen kann
    voraussetzungen: Optional[List[Voraussetzungen]] = None
    einschraenkungzaehler: Optional[List[Geraet]] = None
    """ Liste der Zähler/Geräte, die erforderlich sind, damit dieser Tarif zur Anwendung gelangen kann.
    (Falls keine Zähler angegeben sind, ist der Tarif nicht an das Vorhandensein bestimmter Zähler gebunden.) """
    einschraenkungleistung: Optional[List[Menge]] = None
    """ Die vereinbarte Leistung, die (näherungsweise) abgenommen wird.
    Insbesondere Gastarife können daran gebunden sein, dass die Leistung einer vereinbarten Höhe entspricht. """



"""
Contains Buendelvertrag class and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from datetime import datetime
from typing import List, Optional

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.bo.vertrag import Vertrag
from bo4e.com.unterschrift import Unterschrift
from bo4e.com.vertragskonditionen import Vertragskonditionen
from bo4e.enum.botyp import BoTyp
from bo4e.enum.sparte import Sparte
from bo4e.enum.vertragsart import Vertragsart
from bo4e.enum.vertragsstatus import Vertragsstatus


class Buendelvertrag(Geschaeftsobjekt):
    """
    Abbildung eines Bündelvertrags.
    Es handelt sich hierbei um eine Liste von Einzelverträgen, die in einem Vertragsobjekt gebündelt sind.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Buendelvertrag.svg" type="image/svg+xml"></object>

    .. HINT::
        `Buendelvertrag JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Buendelvertrag.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.BUENDELVERTRAG

    # pylint: disable=duplicate-code
    #: Eine im Verwendungskontext eindeutige Nummer für den Vertrag
    vertragsnummer: str
    #: Hier ist festgelegt, um welche Art von Vertrag es sich handelt. Z.B. Netznutzungvertrag
    vertragsart: Vertragsart
    #: Gibt den Status des Vertrages an
    vertragsstatus: Vertragsstatus
    #: Unterscheidungsmöglichkeiten für die Sparte
    sparte: Sparte
    #: Gibt an, wann der Vertrag beginnt (inklusiv)
    vertragsbeginn: datetime
    #: Gibt an, wann der Vertrag (voraussichtlich) endet oder beendet wurde (exklusiv)
    vertragsende: datetime
    #: Der "erstgenannte" Vertragspartner. In der Regel der Aussteller des Vertrags.
    #: Beispiel: "Vertrag zwischen Vertagspartner 1 ..."
    vertragspartner1: Geschaeftspartner
    #: Der "zweitgenannte" Vertragspartner. In der Regel der Empfänger des Vertrags.
    #: Beispiel "Vertrag zwischen Vertagspartner 1 und Vertragspartner 2"
    vertragspartner2: Geschaeftspartner

    # optional attributes
    #: Die Liste mit den Einzelverträgen zu den Abnahmestellen
    einzelvertraege: Optional[List[Vertrag]] = []
    #: Festlegungen zu Laufzeiten und Kündigungsfristen
    vertragskonditionen: Optional[List[Vertragskonditionen]] = []
    #: Unterzeichner des Vertragspartners1
    unterzeichnervp1: Optional[List[Unterschrift]] = []
    #: Unterzeichner des Vertragspartners2
    unterzeichnervp2: Optional[List[Unterschrift]] = []
    #: Beschreibung zum Vertrag
    beschreibung: Optional[str] = None



from typing import Dict, Optional

import pytest
from py._path.local import LocalPath  # type:ignore[import]
from pydantic import ValidationError

from bo4e.com.adresse import Adresse
from bo4e.enum.landescode import Landescode

# import pydantic

# can be imported by other tests
example_adresse = Adresse(
    ort="Grünwald",
    landescode=Landescode.DE,  # type: ignore[attr-defined]
    hausnummer="27A",
    strasse="Nördliche Münchner Straße",
    postleitzahl="82031",
)


class TestAddress:
    def test_serialization_strasse(self) -> None:
        """
        Test serialization with strasse und hausnummer
        and default value of landescode
        """

        address_test_data = Adresse.parse_file(
            "./tests/test_data/test_data_adresse/test_data_adresse_only_required_fields.json",
            encoding="utf-8",
        )

        address_dict = address_test_data.json(by_alias=True, ensure_ascii=False)

        assert "Nördliche Münchner Straße" in address_dict
        assert "27A" in address_dict
        assert "82031" in address_dict
        assert "DE" in address_dict

        deserialized_address = Adresse.parse_raw(address_dict)

        assert isinstance(deserialized_address, Adresse)
        assert deserialized_address.strasse == "Nördliche Münchner Straße"
        assert deserialized_address.hausnummer == "27A"
        assert deserialized_address.postleitzahl == "82031"
        assert deserialized_address.landescode == Landescode.DE  # type: ignore[attr-defined]

    def test_serialization_only_postfach(self) -> None:
        """Test serialization with postfach"""
        address_test_data = {
            "postleitzahl": "82031",
            "ort": "Grünwald",
            "postfach": "10 64 38",
        }

        a = Adresse(
            postleitzahl=address_test_data["postleitzahl"],
            ort=address_test_data["ort"],
            postfach=address_test_data["postfach"],
        )

        address_json = a.json(by_alias=True, ensure_ascii=False)

        assert "10 64 38" in address_json
        assert "82031" in address_json
        assert "DE" in address_json

        deserialized_address = Adresse.parse_raw(address_json)

        assert isinstance(deserialized_address, Adresse)
        assert deserialized_address.postfach == "10 64 38"
        assert deserialized_address.postleitzahl == "82031"
        assert deserialized_address.landescode == Landescode.DE  # type: ignore[attr-defined]

    def test_serialization_only_required_fields(self) -> None:
        """Test serialization with just postleitzahl und ort"""
        address_test_data = {
            "postleitzahl": "82031",
            "ort": "Grünwald",
        }

        a = Adresse(
            postleitzahl=address_test_data["postleitzahl"],
            ort=address_test_data["ort"],
        )

        address_json = a.json(by_alias=True, ensure_ascii=False)

        assert "Grünwald" in address_json
        assert "82031" in address_json

        deserialized_address = Adresse.parse_raw(address_json)

        assert isinstance(deserialized_address, Adresse)
        assert deserialized_address.ort == "Grünwald"
        assert deserialized_address.postleitzahl == "82031"

    def test_serialization_only_required_fields_landescode_AT(self) -> None:

        address_test_data = Adresse.parse_file(
            "./tests/test_data/test_data_adresse/test_data_adresse_only_required_fields.json",
            encoding="utf-8",
        )
        address_test_data.landescode = Landescode.AT  # type: ignore[attr-defined]

        address_json = address_test_data.json(by_alias=True, ensure_ascii=False)
        deserialized_address = Adresse.parse_raw(address_json)

        assert deserialized_address.landescode == Landescode.AT  # type: ignore[attr-defined]

    def test_deserialization(self) -> None:
        json_string = r"""{"strasse":"Getreidegasse",
                 "hausnummer":"9",
                 "ort":"Salzburg",
                 "postleitzahl":"5020",
                 "landescode":"AT"}"""

        a: Adresse = Adresse.parse_raw(json_string)
        assert a.landescode is Landescode.AT  # type: ignore[attr-defined]

    @pytest.mark.datafiles("./tests/test_data/test_data_adresse/test_data_adresse_missing_plz.json")
    def test_missing_required_attribute(self, datafiles: LocalPath) -> None:
        """
        Test for getting an error message if a required attribute is missing
        """

        with pytest.raises(ValidationError) as excinfo:
            _ = Adresse.parse_file(datafiles / "test_data_adresse_missing_plz.json", encoding="utf-8")

        assert "postleitzahl" in str(excinfo.value)

    @pytest.mark.parametrize(
        "address_test_data, expected",
        [
            pytest.param(
                {
                    "postleitzahl": "82031",
                    "ort": "Grünwald",
                    "strasse": "Nördliche Münchner Straße",
                    "hausnummer": None,
                    "postfach": None,
                },
                'You have to define either "strasse" and "hausnummer" or "postfach".',
                id="Missing hausnummer",
            ),
            pytest.param(
                {
                    "postleitzahl": "82031",
                    "ort": "Grünwald",
                    "strasse": "",
                    "hausnummer": "27A",
                    "postfach": None,
                },
                'You have to define either "strasse" and "hausnummer" or "postfach".',
                id="Missing strasse",
            ),
            pytest.param(
                {
                    "postleitzahl": "82031",
                    "ort": "Grünwald",
                    "strasse": "Nördliche Münchner Straße",
                    "hausnummer": "27A",
                    "postfach": "10 64 38",
                },
                'You have to define either "strasse" and "hausnummer" or "postfach".',
                id="Strasse, hausnummer and postfach",
            ),
            pytest.param(
                {
                    "postleitzahl": "82031",
                    "ort": "Grünwald",
                    "strasse": None,
                    "hausnummer": "27A",
                    "postfach": "10 64 38",
                },
                'You have to define either "strasse" and "hausnummer" or "postfach".',
                id="Postfach and hausnummer",
            ),
        ],
    )
    def test_strasse_xor_postfach(self, address_test_data: Dict[str, Optional[str]], expected: str) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Adresse(
                postleitzahl=address_test_data["postleitzahl"],  # type: ignore[arg-type]
                ort=address_test_data["ort"],  # type: ignore[arg-type]
                strasse=address_test_data["strasse"],
                hausnummer=address_test_data["hausnummer"],
                postfach=address_test_data["postfach"],
            )
        assert expected in str(excinfo.value)

    def test_serialization_of_non_german_address(self) -> None:
        """
        Minimal working example
        :return:
        """
        a = Adresse(
            postleitzahl="6413", ort="Wildermieming", strasse="Gerhardhof", hausnummer="1", landescode=Landescode.AT  # type: ignore[attr-defined]
        )
        assert a.landescode == Landescode.AT  # type: ignore[attr-defined]
        serialized_address = a.json(by_alias=True, ensure_ascii=False)
        assert '"AT"' in serialized_address
        deserialized_address = Adresse.parse_raw(serialized_address)
        assert deserialized_address.landescode == Landescode.AT  # type: ignore[attr-defined]



"""
Contains Unterschrift class
and corresponding marshmallow schema for de-/serialization
"""
from datetime import datetime
from typing import Optional

from bo4e.com.com import COM

# pylint: disable=too-few-public-methods


class Unterschrift(COM):
    """
    Modellierung einer Unterschrift, z.B. für Verträge, Angebote etc.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Unterschrift.svg" type="image/svg+xml"></object>

    .. HINT::
        `Unterschrift JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Unterschrift.json>`_

    """

    # required attributes
    #: Name des Unterschreibers
    name: str

    # optional attributes
    ort: Optional[str] = None  #: Ort, an dem die Unterschrift geleistet wird
    datum: Optional[datetime] = None  #: Datum der Unterschrift



"""
Contains PreisblattHardware class and corresponding marshmallow schema for de-/serialization
"""
from typing import List, Optional

from bo4e.bo.preisblatt import Preisblatt
from bo4e.com.geraeteeigenschaften import Geraeteeigenschaften
from bo4e.enum.bilanzierungsmethode import Bilanzierungsmethode
from bo4e.enum.botyp import BoTyp
from bo4e.enum.dienstleistungstyp import Dienstleistungstyp
from bo4e.enum.netzebene import Netzebene

# pylint: disable=too-few-public-methods


class PreisblattHardware(Preisblatt):
    """
    Variante des Preisblattmodells zur Abbildung der Preise für zusätzliche Hardware

    .. raw:: html

        <object data="../_static/images/bo4e/bo/PreisblattHardware.svg" type="image/svg+xml"></object>

    .. HINT::
        `PreisblattHardware JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/PreisblattHardware.json>`_

    """

    bo_typ: BoTyp = BoTyp.PREISBLATTHARDWARE
    # required attributes (additional to those of Preisblatt)
    #: Die Preise gelten für Marktlokationen der angebebenen Bilanzierungsmethode
    bilanzierungsmethode: Bilanzierungsmethode
    #: Die Preise gelten für Messlokationen in der angebebenen Netzebene
    messebene: Netzebene

    #: Der Preis betriftt das hier angegebene Gerät, z.B. ein Tarifschaltgerät
    basisgeraet: Geraeteeigenschaften

    # optional attributes
    #: Im Preis sind die hier angegebenen Dienstleistungen enthalten, z.B. Jährliche Ablesung
    inklusive_dienstleistungen: Optional[List[Dienstleistungstyp]] = None

    #: Im Preis sind die hier angegebenen Geräte mit enthalten, z.B. ein Wandler
    inklusive_geraete: Optional[List[Geraeteeigenschaften]] = None



"""
This script is run in the tox 'json_schemas' environment.
It creates json schema files as described in the README.md in the same directory.
"""

import importlib
import inspect
import pathlib
import pkgutil

pkgs = ["bo", "com"]
for pkg in pkgs:
    modls = [
        name for _, name, _ in pkgutil.iter_modules([str(pathlib.Path(__file__).parent.parent / "src" / "bo4e" / pkg)])
    ]
    for x in modls:
        modl_name = f"bo4e.{pkg}.{x}"
        modl = importlib.import_module(modl_name)
        # pylint: disable=cell-var-from-loop
        cls_list = inspect.getmembers(modl, lambda member: inspect.isclass(member) and member.__module__ == modl_name)
        for name, cls in cls_list:
            this_directory = pathlib.Path(__file__).parent.absolute()
            file_path = this_directory / pkg / (name + ".json")  # pylint:disable=invalid-name
            with open(file_path, "w+", encoding="utf-8") as json_schema_file:
                json_schema_file.write(cls.schema_json(ensure_ascii=False, sort_keys=True, indent=4))



"""
Contains Tarifinfo class
and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-many-instance-attributes, too-few-public-methods
# pylint: disable=no-name-in-module
from datetime import datetime
from typing import Optional

from pydantic import conlist

from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.bo.marktteilnehmer import Marktteilnehmer
from bo4e.com.energiemix import Energiemix
from bo4e.com.vertragskonditionen import Vertragskonditionen
from bo4e.com.zeitraum import Zeitraum
from bo4e.enum.botyp import BoTyp
from bo4e.enum.kundentyp import Kundentyp
from bo4e.enum.sparte import Sparte
from bo4e.enum.tarifart import Tarifart
from bo4e.enum.tarifmerkmal import Tarifmerkmal
from bo4e.enum.tariftyp import Tariftyp


class Tarifinfo(Geschaeftsobjekt):
    """
    Das BO Tarifinfo liefert die Merkmale, die einen Endkundentarif identifizierbar machen.
    Dieses BO dient als Basis für weitere BOs mit erweiterten Anwendungsmöglichkeiten.

    .. raw:: html

        <object data="../_static/images/bo4e/bo/Tarifinfo.svg" type="image/svg+xml"></object>

    .. HINT::
        `Tarifinfo JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/bo/Tarifinfo.json>`_

    """

    # required attributes
    bo_typ: BoTyp = BoTyp.TARIFINFO
    #: Name des Tarifs
    bezeichnung: str
    #: Der Name des Marktpartners, der den Tarif anbietet
    anbietername: str
    #: Strom oder Gas, etc.
    sparte: Sparte
    #: Kundentypen für den der Tarif gilt, z.B. Privatkunden
    kundentypen: conlist(Kundentyp, min_items=1)  # type: ignore[valid-type]
    #: Die Art des Tarifes, z.B. Eintarif oder Mehrtarif
    tarifart: Tarifart
    #: Hinweis auf den Tariftyp, z.B. Grundversorgung oder Sondertarif
    tariftyp: Tariftyp
    #: Weitere Merkmale des Tarifs, z.B. Festpreis oder Vorkasse
    tarifmerkmale: conlist(Tarifmerkmal, min_items=1)  # type: ignore[valid-type]
    #: Der Marktteilnehmer (Lieferant), der diesen Tarif anbietet
    anbieter: Marktteilnehmer

    # optional attributes
    #: Internetseite auf dem der Tarif zu finden ist
    website: Optional[str] = None
    #: Freitext
    bemerkung: Optional[str] = None

    #: Angabe, in welchem Zeitraum der Tarif gültig ist
    zeitliche_gueltigkeit: Optional[Zeitraum] = None
    #: Der Energiemix, der für diesen Tarif gilt
    energiemix: Optional[Energiemix] = None
    #: Mindestlaufzeiten und Kündigungsfristen zusammengefasst
    vertragskonditionen: Optional[Vertragskonditionen] = None
    anwendung_von: Optional[datetime] = None
    """
    Angabe des inklusiven Zeitpunkts, ab dem der Tarif bzw. der Preis angewendet und abgerechnet wird,
    z.B. "2021-07-20T18:31:48Z"
    """



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Netzebene(StrEnum):
    """
    Auflistung möglicher Netzebenen innerhalb der Energiearten Strom und Gas.
    """

    NSP = "NSP"  #: Niederspannung; Strom
    MSP = "MSP"  #: Mittelspannung; Strom
    HSP = "HSP"  #: Hochspannung; Strom
    HSS = "HSS"  #: Hoechstspannung; Strom
    MSP_NSP_UMSP = "MSP_NSP_UMSP"  #: MS/NS Umspannung; Strom
    HSP_MSP_UMSP = "HSP_MSP_UMSP"  #: HS/MS Umspannung; Strom
    HSS_HSP_UMSP = "HSS_HSP_UMSP"  #: HOES/HS Umspannung; Strom
    HD = "HD"  #: Hochdruck; Gas
    MD = "MD"  #: Mitteldruck; Gas
    ND = "ND"  #: Niederdruck; Gas



from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.energieherkunft import Energieherkunft
from bo4e.com.energiemix import Energiemix
from bo4e.enum.erzeugungsart import Erzeugungsart
from bo4e.enum.oekolabel import Oekolabel
from bo4e.enum.oekozertifikat import Oekozertifikat
from bo4e.enum.sparte import Sparte
from tests.serialization_helper import assert_serialization_roundtrip

example_energiemix = Energiemix(
    energiemixnummer=2,
    energieart=Sparte.STROM,
    bezeichnung="foo",
    gueltigkeitsjahr=2021,
    anteil=[
        Energieherkunft(
            erzeugungsart=Erzeugungsart.BIOGAS,
            anteil_prozent=Decimal(40),
        ),
    ],
)


class TestEnergiemix:
    @pytest.mark.parametrize(
        "energiemix, expected_json_dict",
        [
            pytest.param(
                example_energiemix,
                {
                    "energiemixnummer": 2,
                    "energieart": Sparte.STROM,
                    "bezeichnung": "foo",
                    "gueltigkeitsjahr": 2021,
                    "anteil": [
                        {
                            "erzeugungsart": Erzeugungsart.BIOGAS,
                            "anteilProzent": Decimal("40"),
                        }
                    ],
                    "oekolabel": [],
                    "bemerkung": None,
                    "co2Emission": None,
                    "atommuell": None,
                    "website": None,
                    "oekozertifikate": [],
                    "oekoTopTen": None,
                },
                id="only required attributes",
            ),
            pytest.param(
                Energiemix(
                    energiemixnummer=2,
                    energieart=Sparte.STROM,
                    bezeichnung="foo",
                    gueltigkeitsjahr=2021,
                    anteil=[
                        Energieherkunft(
                            erzeugungsart=Erzeugungsart.BIOGAS,
                            anteil_prozent=Decimal(40),
                        ),
                        Energieherkunft(
                            erzeugungsart=Erzeugungsart.GEOTHERMIE,
                            anteil_prozent=Decimal(60),
                        ),
                    ],
                    oekolabel=[
                        Oekolabel.GASGREEN,
                        Oekolabel.GRUENER_STROM_GOLD,
                    ],
                    bemerkung="bar",
                    co2_emission=Decimal(40),
                    atommuell=Decimal(5),
                    website="foobar.de",
                    oekozertifikate=[Oekozertifikat.FRAUNHOFER, Oekozertifikat.FREIBERG],
                    oeko_top_ten=True,
                ),
                {
                    "energiemixnummer": 2,
                    "energieart": Sparte.STROM,
                    "bezeichnung": "foo",
                    "gueltigkeitsjahr": 2021,
                    "anteil": [
                        {
                            "erzeugungsart": Erzeugungsart.BIOGAS,
                            "anteilProzent": Decimal("40"),
                        },
                        {
                            "erzeugungsart": Erzeugungsart.GEOTHERMIE,
                            "anteilProzent": Decimal("60"),
                        },
                    ],
                    "oekolabel": ["GASGREEN", "GRUENER_STROM_GOLD"],
                    "bemerkung": "bar",
                    "co2Emission": Decimal("40"),
                    "atommuell": Decimal("5"),
                    "website": "foobar.de",
                    "oekozertifikate": ["FRAUNHOFER", "FREIBERG"],
                    "oekoTopTen": True,
                },
                id="required and optional attributes",
            ),
        ],
    )
    def test_energiemix_serialization_roundtrip(
        self, energiemix: Energiemix, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of Energiehermix with minimal attributes.
        """
        assert_serialization_roundtrip(energiemix, expected_json_dict)

    def test_energiemix_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Energiemix()  # type: ignore[call-arg]

        assert "5 validation errors" in str(excinfo.value)

    def test_energiemix_anteil_required(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Energiemix(
                energiemixnummer=2,
                energieart=Sparte.STROM,
                bezeichnung="foo",
                gueltigkeitsjahr=2021,
                anteil=[],
            )

        assert "1 validation error" in str(excinfo.value)
        assert "ensure this value has at least 1 item" in str(excinfo.value)



from decimal import Decimal
from typing import Tuple

import pytest
from pydantic import ValidationError

from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.bo.marktlokation import Marktlokation
from bo4e.com.adresse import Adresse
from bo4e.com.geokoordinaten import Geokoordinaten
from bo4e.enum.anrede import Anrede
from bo4e.enum.bilanzierungsmethode import Bilanzierungsmethode
from bo4e.enum.botyp import BoTyp
from bo4e.enum.energierichtung import Energierichtung
from bo4e.enum.geschaeftspartnerrolle import Geschaeftspartnerrolle
from bo4e.enum.kontaktart import Kontaktart
from bo4e.enum.kundentyp import Kundentyp
from bo4e.enum.netzebene import Netzebene
from bo4e.enum.sparte import Sparte


class TestMaLo:
    def test_serialisation_only_required_attributes(self) -> None:
        """
        Test serialisation of Marktlokation only with required attributes
        """
        malo = Marktlokation(
            marktlokations_id="51238696781",
            sparte=Sparte.GAS,
            lokationsadresse=Adresse(postleitzahl="04177", ort="Leipzig", hausnummer="1", strasse="Jahnalle"),
            energierichtung=Energierichtung.EINSP,
            bilanzierungsmethode=Bilanzierungsmethode.PAUSCHAL,
            netzebene=Netzebene.NSP,
        )
        assert malo.versionstruktur == "2", "versionstruktur was not automatically set"
        assert malo.bo_typ is BoTyp.MARKTLOKATION, "boTyp was not automatically set"

        json_string = malo.json(by_alias=True, ensure_ascii=False)

        assert "boTyp" in json_string, "No camel case serialization"
        assert "marktlokationsId" in json_string, "No camel case serialization"

        deserialized_malo: Marktlokation = Marktlokation.parse_raw(json_string)

        # check that `deserialized_malo.marktlokations_id` and `malo.marktlokations_id` have the same value
        # but are **not** the same object.
        assert deserialized_malo.marktlokations_id == malo.marktlokations_id
        assert deserialized_malo.marktlokations_id is not malo.marktlokations_id
        assert deserialized_malo.bo_typ is BoTyp.MARKTLOKATION

    def test_serialization_required_and_optional_attributes(self) -> None:
        """
        Test serialisation of Marktlokation with required attributes and optional attributes
        """
        gp = Geschaeftspartner(
            anrede=Anrede.FRAU,
            name1="von Sinnen",
            name2="Helga",
            name3=None,
            gewerbekennzeichnung=True,
            hrnummer="HRB 254466",
            amtsgericht="Amtsgericht München",
            kontaktweg=[Kontaktart.E_MAIL],
            umsatzsteuer_id="DE267311963",
            glaeubiger_id="DE98ZZZ09999999999",
            e_mail_adresse="test@bo4e.de",
            website="bo4e.de",
            geschaeftspartnerrolle=[Geschaeftspartnerrolle.DIENSTLEISTER],
            partneradresse=Adresse(
                postleitzahl="82031",
                ort="Grünwald",
                strasse="Nördliche Münchner Straße",
                hausnummer="27A",
            ),
        )
        malo = Marktlokation(
            marktlokations_id="51238696781",
            sparte=Sparte.GAS,
            lokationsadresse=Adresse(postleitzahl="04177", ort="Leipzig", hausnummer="1", strasse="Jahnalle"),
            energierichtung=Energierichtung.EINSP,
            bilanzierungsmethode=Bilanzierungsmethode.PAUSCHAL,
            unterbrechbar=True,  # optional attribute
            netzebene=Netzebene.NSP,
            endkunde=gp,
            kundengruppen=[Kundentyp.GEWERBE, Kundentyp.PRIVAT],
        )
        assert malo.versionstruktur == "2", "versionstruktur was not automatically set"
        assert malo.bo_typ == BoTyp.MARKTLOKATION, "boTyp was not automatically set"

        json_string = malo.json(by_alias=True, ensure_ascii=False)

        assert "boTyp" in json_string, "No camel case serialization"
        assert "marktlokationsId" in json_string, "No camel case serialization"

        deserialized_malo: Marktlokation = Marktlokation.parse_raw(json_string)

        assert deserialized_malo.marktlokations_id == malo.marktlokations_id
        assert deserialized_malo.marktlokations_id is not malo.marktlokations_id
        assert deserialized_malo.bo_typ is BoTyp.MARKTLOKATION
        assert deserialized_malo.endkunde == gp

    def test_missing_required_fields(self) -> None:
        """
        Test that the required attributes are checked in the deserialization.
        Therefore the required attribute `marktlokations_id` is removed in the test data.
        """
        invalid_json_string = """{
            "katasterinformation": null,
            "boTyp": "MARKTLOKATION",
            "endkunde": null,
            "sparte": "GAS",
            "zugehoerigeMesslokation": null,
            "verbrauchsart": null,
            "externeReferenzen": null,
            "gebietstyp": null,
            "grundversorgercodenr": null,
            "netzebene": "NSP",
            "netzbetreibercodenr": null,
            "netzgebietsnr": null,
            "lokationsadresse": {
                "ort": "Leipzig",
                "strasse": "Jahnalle",
                "adresszusatz": null,
                "coErgaenzung": null,
                "landescode": "DE",
                "postfach": null,
                "hausnummer": "1",
                "postleitzahl": "04177"
                },
            "unterbrechbar": null,
            "gasqualitaet": null,
            "bilanzierungsgebiet": null,
            "geoadresse": null,
            "bilanzierungsmethode": "PAUSCHAL",
            "versionstruktur": "2",
            "energierichtung": "EINSP"
            }"""

        with pytest.raises(ValidationError) as excinfo:
            Marktlokation.parse_raw(invalid_json_string)

        assert "1 validation error" in str(excinfo.value)
        assert "marktlokationsId" in str(excinfo.value)
        assert "field required" in str(excinfo.value)

    def test_address_validation(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Marktlokation(
                marktlokations_id="51238696781",
                sparte=Sparte.GAS,
                lokationsadresse=Adresse(
                    postleitzahl="04177",
                    ort="Leipzig",
                    hausnummer="1",
                    strasse="Jahnalle",
                ),
                energierichtung=Energierichtung.EINSP,
                bilanzierungsmethode=Bilanzierungsmethode.PAUSCHAL,
                unterbrechbar=True,  # optional attribute
                netzebene=Netzebene.NSP,
                geoadresse=Geokoordinaten(
                    breitengrad=Decimal("52"),
                    laengengrad=Decimal("9"),
                ),
                katasterinformation=None,
            )

        assert "No or more than one address information is given." in str(excinfo.value)

    @pytest.mark.parametrize(
        "malo_id_valid",
        [
            ("51238696781", True),
            ("41373559241", True),
            ("56789012345", True),
            ("52935155442", True),
            ("12345678910", False),
            ("asdasd", False),
            ("   ", False),
            ("  asdasdasd ", False),
            ("keine malo id", False),
            (None, False),
            ("", False),
        ],
    )
    def test_id_validation(self, malo_id_valid: Tuple[str, bool]) -> None:
        def _instantiate_malo(malo_id: str) -> None:
            _ = Marktlokation(
                marktlokations_id=malo_id,
                sparte=Sparte.GAS,
                lokationsadresse=Adresse(
                    postleitzahl="82031",
                    ort="Grünwald",
                    hausnummer="27A",
                    strasse="Nördliche Münchner Straße",
                ),
                energierichtung=Energierichtung.EINSP,
                bilanzierungsmethode=Bilanzierungsmethode.PAUSCHAL,
                unterbrechbar=True,
                netzebene=Netzebene.NSP,
            )

        if not malo_id_valid[1]:
            with pytest.raises(ValidationError):
                _instantiate_malo(malo_id_valid[0])
        else:
            _instantiate_malo(malo_id_valid[0])



"""
Contains Kostenposition and corresponding marshmallow schema for de-/serialization
"""
from datetime import datetime
from typing import Any, Dict, Optional

# pylint: disable=too-few-public-methods, too-many-instance-attributes
from pydantic import validator

from bo4e.com.betrag import Betrag
from bo4e.com.com import COM
from bo4e.com.menge import Menge
from bo4e.com.preis import Preis
from bo4e.validators import check_bis_is_later_than_von


class Kostenposition(COM):
    """
    Diese Komponente wird zur Übertagung der Details zu einer Kostenposition verwendet.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Kostenposition.svg" type="image/svg+xml"></object>

    .. HINT::
        `Kostenposition JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Kostenposition.json>`_

    """

    # required attributes
    #: Ein Titel für die Zeile. Hier kann z.B. der Netzbetreiber eingetragen werden, wenn es sich um Netzkosten handelt.
    positionstitel: str

    betrag_kostenposition: Betrag
    """Der errechnete Gesamtbetrag der Position als Ergebnis der Berechnung <Menge * Einzelpreis> oder
    <Einzelpreis / (Anzahl Tage Jahr) * zeitmenge>"""
    # todo: validate above calculation, see https://github.com/Hochfrequenz/BO4E-python/issues/282

    #: Bezeichnung für den Artikel für den die Kosten ermittelt wurden. Beispiel: Arbeitspreis HT
    artikelbezeichnung: str

    #: Der Preis für eine Einheit. Beispiele: 5,8200 ct/kWh oder 55 €/Jahr.
    einzelpreis: Preis

    # optional attributes
    #: inklusiver von-Zeitpunkt der Kostenzeitscheibe
    von: Optional[datetime] = None
    #: exklusiver bis-Zeitpunkt der Kostenzeitscheibe
    bis: Optional[datetime] = None
    _bis_check = validator("bis", always=True, allow_reuse=True)(check_bis_is_later_than_von)

    #: Die Menge, die in die Kostenberechnung eingeflossen ist. Beispiel: 3.660 kWh
    menge: Optional[Menge] = None

    zeitmenge: Optional[Menge] = None
    """
    Wenn es einen zeitbasierten Preis gibt (z.B. €/Jahr), dann ist hier die Menge angegeben mit der die Kosten berechnet
    wurden. Z.B. 138 Tage.
    """

    #: Detaillierung des Artikels (optional). Beispiel: 'Drehstromzähler'
    artikeldetail: Optional[str] = None

    @staticmethod
    def _get_inclusive_start(values: Dict[str, Any]) -> Optional[datetime]:
        return values["von"]

    # @staticmethod
    # def _get_exclusive_end(values) -> Optional[datetime]:
    #     return values['bis']



from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.energieherkunft import Energieherkunft
from bo4e.enum.erzeugungsart import Erzeugungsart
from tests.serialization_helper import assert_serialization_roundtrip

# this can be imported by other tests
example_energieherkunft = Energieherkunft(erzeugungsart=Erzeugungsart.BIOMASSE, anteil_prozent=Decimal(25.5))


class TestEnergieherkunft:
    @pytest.mark.parametrize(
        "energieherkunft, expected_json_dict",
        [
            pytest.param(
                example_energieherkunft,
                {
                    "erzeugungsart": Erzeugungsart.BIOMASSE,
                    "anteilProzent": Decimal("25.5"),
                },
            ),
        ],
    )
    def test_energieherkunft_required_attributes(
        self, energieherkunft: Energieherkunft, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of Energieherkunft with minimal attributes.
        """
        assert_serialization_roundtrip(energieherkunft, expected_json_dict)

    def test_energieherkunft_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Energieherkunft()  # type: ignore[call-arg]

        assert "2 validation errors" in str(excinfo.value)

    @pytest.mark.parametrize(
        "failing_percentage",
        [
            pytest.param(100.1),
            pytest.param(-2),
        ],
    )
    def test_energieherkunft_failing_validation(self, failing_percentage: float) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = (Energieherkunft(erzeugungsart=Erzeugungsart.BIOMASSE, anteil_prozent=Decimal(failing_percentage)),)

        assert "1 validation error" in str(excinfo.value)
        assert "ensure this value is less than 100" in str(
            excinfo.value
        ) or "ensure this value is greater than 0" in str(excinfo.value)



from datetime import datetime, timezone

import pytest
from pydantic import ValidationError

from bo4e.bo.tarif import Tarif
from bo4e.enum.kundentyp import Kundentyp
from bo4e.enum.sparte import Sparte
from bo4e.enum.tarifart import Tarifart
from bo4e.enum.tarifmerkmal import Tarifmerkmal
from bo4e.enum.tariftyp import Tariftyp
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_aufabschlagregional import example_aufabschlagregional
from tests.test_energiemix import example_energiemix
from tests.test_marktteilnehmer import example_marktteilnehmer
from tests.test_preisgarantie import example_preisgarantie
from tests.test_regionaletarifpreisposition import example_regionale_tarifpreisposition
from tests.test_tarifberechnungsparameter import example_tarifberechnungsparameter
from tests.test_tarifeinschraenkung import example_tarifeinschraenkung
from tests.test_tarifpreispositionproort import example_tarifpreispositionproort
from tests.test_vertragskonditionen import example_vertragskonditionen
from tests.test_zeitraum import example_zeitraum


class TestTarif:
    @pytest.mark.parametrize(
        "tarif",
        [
            pytest.param(
                Tarif(
                    preisstand=datetime(2022, 2, 1, 0, 0, 0, tzinfo=timezone.utc),
                    berechnungsparameter=example_tarifberechnungsparameter,
                    tarif_auf_abschlaege=[example_aufabschlagregional],
                    tarifpreise=[example_tarifpreispositionproort],
                    preisgarantie=example_preisgarantie,
                    tarifeinschraenkung=example_tarifeinschraenkung,
                    # below are the attributes of tarifinfo
                    bezeichnung="foo",
                    anbietername="der beste stromanbieter",
                    sparte=Sparte.STROM,
                    kundentypen=[Kundentyp.PRIVAT, Kundentyp.GEWERBE],
                    tarifart=Tarifart.MEHRTARIF,
                    tariftyp=Tariftyp.GRUND_ERSATZVERSORGUNG,
                    tarifmerkmale=[Tarifmerkmal.HEIZSTROM],
                    website="https://foo.inv",
                    bemerkung="super billig aber auch super dreckig",
                    vertragskonditionen=example_vertragskonditionen,
                    zeitliche_gueltigkeit=example_zeitraum,
                    energiemix=example_energiemix,
                    anbieter=example_marktteilnehmer,
                ),
                id="required and optional attributes",
            ),
            pytest.param(
                Tarif(
                    preisstand=datetime(2022, 2, 1, 0, 0, 0, tzinfo=timezone.utc),
                    berechnungsparameter=example_tarifberechnungsparameter,
                    tarifpreise=[example_tarifpreispositionproort],
                    # below are the attributes of tarifinfo
                    bezeichnung="foo",
                    anbietername="der beste stromanbieter",
                    sparte=Sparte.STROM,
                    kundentypen=[Kundentyp.PRIVAT, Kundentyp.GEWERBE],
                    tarifart=Tarifart.MEHRTARIF,
                    tariftyp=Tariftyp.GRUND_ERSATZVERSORGUNG,
                    tarifmerkmale=[Tarifmerkmal.HEIZSTROM],
                    anbieter=example_marktteilnehmer,
                ),
                id="only required attributes",
            ),
        ],
    )
    def test_serialization_roundtrip(self, tarif: Tarif) -> None:
        """
        Test de-/serialisation
        """
        assert_serialization_roundtrip(tarif)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Tarif()  # type: ignore[call-arg]
        assert "11 validation error" in str(excinfo.value)  # 3 from Tarif + 8 from tarifinfo



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Rollencodetyp(StrEnum):
    """
    Gibt den Codetyp einer Rolle, beispielsweise einer Marktrolle, an.
    """

    BDEW = "BDEW"  #: Bundesverband der Energie- u. Wasserwirtschaft
    DVG = "DVG"  #: Deutscher Verein des Gas- und Wasserfaches
    GLN = "GLN"  #: Global Location Number



import json
from datetime import datetime
from decimal import Decimal

import pytest
from pydantic import ValidationError

from bo4e.bo.messlokation import Messlokation
from bo4e.bo.zaehler import Zaehler
from bo4e.com.adresse import Adresse
from bo4e.com.dienstleistung import Dienstleistung
from bo4e.com.externereferenz import ExterneReferenz
from bo4e.com.geokoordinaten import Geokoordinaten
from bo4e.com.hardware import Hardware
from bo4e.com.katasteradresse import Katasteradresse
from bo4e.com.zaehlwerk import Zaehlwerk
from bo4e.enum.botyp import BoTyp
from bo4e.enum.dienstleistungstyp import Dienstleistungstyp
from bo4e.enum.energierichtung import Energierichtung
from bo4e.enum.geraetetyp import Geraetetyp
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.netzebene import Netzebene
from bo4e.enum.sparte import Sparte
from bo4e.enum.tarifart import Tarifart
from bo4e.enum.zaehlerauspraegung import Zaehlerauspraegung
from bo4e.enum.zaehlertyp import Zaehlertyp


class TestMeLo:
    def test_serialisation_only_required_attributes(self) -> None:
        """
        Test serialisation of Messlokation only with required attributes
        """
        melo = Messlokation(
            messlokations_id="DE00056266802AO6G56M11SN51G21M24S",
            sparte=Sparte.STROM,
        )
        assert melo.versionstruktur == "2", "versionstruktur was not automatically set"
        assert melo.bo_typ is BoTyp.MESSLOKATION, "boTyp was not automatically set"

        json_string = melo.json(by_alias=True, ensure_ascii=False)
        json_dict = json.loads(json_string)

        assert "boTyp" in json_dict, "No camel case serialization"
        assert "messlokationsId" in json_dict, "No camel case serialization"

        deserialized_melo: Messlokation = Messlokation.parse_raw(json_string)

        # check that `deserialized_malo.marktlokations_id` and `malo.marktlokations_id` have the same value
        # but are **not** the same object.
        assert deserialized_melo.messlokations_id == melo.messlokations_id
        assert deserialized_melo.messlokations_id is not melo.messlokations_id
        assert deserialized_melo.bo_typ is BoTyp.MESSLOKATION

    def test_serialization_required_and_optional_attributes(self) -> None:
        """
        Test serialisation of Messlokation with required attributes and optional attributes
        """

        melo = Messlokation(
            # required attributes
            messlokations_id="DE00056266802AO6G56M11SN51G21M24S",
            sparte=Sparte.STROM,
            # optional attributes
            netzebene_messung=Netzebene.MSP,
            messgebietnr="664073",
            geraete=[
                Hardware(geraetetyp=Geraetetyp.INTELLIGENTES_MESSYSTEM, bezeichnung="intelligentes Messsystem"),
                Hardware(geraetetyp=Geraetetyp.MODEM, bezeichnung="56k Modem"),
            ],
            messdienstleistung=[
                Dienstleistung(
                    dienstleistungstyp=Dienstleistungstyp.AUSLESUNG_TAEGLICH_FERNAUSLESUNG,
                    bezeichnung="fernauslesung_taeglich",
                ),
                Dienstleistung(
                    dienstleistungstyp=Dienstleistungstyp.ENTSPERRUNG,
                    bezeichnung="entsperrung",
                ),
            ],
            messlokationszaehler=[
                Zaehler(
                    zaehlernummer="000111222",
                    sparte=Sparte.STROM,
                    zaehlerauspraegung=Zaehlerauspraegung.EINRICHTUNGSZAEHLER,
                    zaehlwerke=[
                        Zaehlwerk(
                            zaehlwerk_id="98765",
                            einheit=Mengeneinheit.KW,
                            richtung=Energierichtung.EINSP,
                            bezeichnung="my zaehlwerk",
                            obis_kennzahl="1-0:1.8.1",
                            wandlerfaktor=Decimal(0.95),
                        )
                    ],
                    zaehlertyp=Zaehlertyp.DREHSTROMZAEHLER,
                    tarifart=Tarifart.ZWEITARIF,
                    zaehlerkonstante=Decimal(0.9),
                    eichung_bis=datetime(2022, 1, 1, 0, 0, 0),
                    externe_referenzen=[ExterneReferenz(ex_ref_name="zaehler im anderen system", ex_ref_wert="7890")],
                    letzte_eichung=datetime(2019, 6, 30, 0, 0, 0),
                )
            ],
            grundzustaendiger_msb_codenr="9910125000002",
            messadresse=Adresse(postleitzahl="04177", ort="Leipzig", hausnummer="1", strasse="Jahnalle"),
        )
        assert melo.versionstruktur == "2", "versionstruktur was not automatically set"
        assert melo.bo_typ == BoTyp.MESSLOKATION, "boTyp was not automatically set"

        json_string = melo.json(by_alias=True, ensure_ascii=False)
        json_dict = json.loads(json_string)

        assert "boTyp" in json_dict, "No camel case serialization"
        assert "messlokationsId" in json_dict, "No camel case serialization"

        deserialized_melo: Messlokation = Messlokation.parse_raw(json_string)

        assert deserialized_melo.messlokations_id == melo.messlokations_id
        assert deserialized_melo.messlokations_id is not melo.messlokations_id
        assert deserialized_melo.bo_typ is BoTyp.MESSLOKATION

    def test_missing_required_fields(self) -> None:
        """
        Test that the required attributes are checked in the deserialization.
        Therefore the required attribute `messlokations_id` is removed in the test data.
        """
        invalid_json_string = """
            {
                "boTyp": "MESSLOKATION",
                "messlokationszaehler": null,
                "katasterinformation": null,
                "messgebietnr": "664073",
                "messadresse":
                    {
                        "strasse": "Jahnalle",
                        "hausnummer": "1",
                        "postfach": null,
                        "postleitzahl": "04177",
                        "landescode": "DE",
                        "coErgaenzung": null,
                        "adresszusatz": null,
                        "ort": "Leipzig"
                    },
                "sparte": "STROM",
                "netzebeneMessung": "MSP",
                "grundzustaendigerMsbCodenr": null,
                "geraete": null,
                "externeReferenzen": [],
                "messdienstleistung": null,
                "geoadresse": null,
                "versionstruktur": "2",
                "grundzustaendigerMsbimCodenr": null
                }
                """

        with pytest.raises(ValidationError) as excinfo:
            Messlokation.parse_raw(invalid_json_string)

        assert "messlokationsId" in str(excinfo.value)

    def test_address_validation(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Messlokation(
                messlokations_id="DE00056266802AO6G56M11SN51G21M24S",
                sparte=Sparte.STROM,
                netzebene_messung=Netzebene.MSP,
                messadresse=Adresse(postleitzahl="04177", ort="Leipzig", hausnummer="1", strasse="Jahnalle"),
                geoadresse=Geokoordinaten(
                    breitengrad=Decimal(52.52149200439453),
                    laengengrad=Decimal(13.404866218566895),
                ),
                katasterinformation=Katasteradresse(gemarkung_flur="hello", flurstueck="world"),
            )

        assert "More than one address information is given." in str(excinfo.value)

    def test_grundzustaendiger_x_codenr_validation(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Messlokation(
                messlokations_id="DE00056266802AO6G56M11SN51G21M24S",
                sparte=Sparte.STROM,
                netzebene_messung=Netzebene.MSP,
                grundzustaendiger_msb_codenr="9904768000008",
                grundzustaendiger_msbim_codenr="test",
            )

        assert "More than one codenr is given." in str(excinfo.value)

    @pytest.mark.parametrize(
        "melo_id, is_valid",
        [
            ("DE00056266802AO6G56M11SN51G21M24S", True),
            ("FR00056266802AO6G56M11SN51G21M24S", True),
            ("XX00056266802AO6G56M11SN51G21M24S", False),
            ("0000056266802AO6G56M11SN51G21M24S", False),
            ("asdasd", False),
            ("   ", False),
            ("  asdasdasd ", False),
            ("keine melo id", False),
            (None, False),
            ("", False),
        ],
    )
    def test_id_validation(self, melo_id: str, is_valid: bool) -> None:
        def _instantiate_melo(melo_id: str) -> None:
            _ = Messlokation(
                messlokations_id=melo_id,
                sparte=Sparte.STROM,
                netzebene_messung=Netzebene.MSP,
            )

        if not is_valid:
            with pytest.raises(ValidationError):
                _instantiate_melo(melo_id)
        else:
            _instantiate_melo(melo_id)



"""
Contains TarifpreispositionProOrt class
and corresponding marshmallow schema for de-/serialization
"""

# pylint: disable=too-few-public-methods
# pylint: disable=no-name-in-module
from pydantic import conlist, constr

from bo4e.com.com import COM
from bo4e.com.tarifpreisstaffelproort import TarifpreisstaffelProOrt


class TarifpreispositionProOrt(COM):
    """
    Mit dieser Komponente können Tarifpreise verschiedener Typen abgebildet werden

    .. raw:: html

        <object data="../_static/images/bo4e/com/TarifpreispositionProOrt.svg" type="image/svg+xml"></object>

    .. HINT::
        `TarifpreispositionProOrt JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/TarifpreispositionProOrt.json>`_

    """

    # required attributes
    #: Postleitzahl des Ortes für den der Preis gilt
    postleitzahl: constr(strict=True, regex=r"^\d{5}$")  # type: ignore[valid-type]
    #: Ort für den der Preis gilt
    ort: str
    #: ene't-Netznummer des Netzes in dem der Preis gilt
    netznr: str
    # Hier sind die Staffeln mit ihren Preisenangaben definiert
    preisstaffeln: conlist(TarifpreisstaffelProOrt, min_items=1)  # type: ignore[valid-type]
    # there are no optional attributes



"""
Contains Preisstaffel and corresponding marshmallow schema for de-/serialization
"""

from decimal import Decimal
from typing import Optional

from bo4e.com.com import COM
from bo4e.com.sigmoidparameter import Sigmoidparameter

# pylint: disable=too-few-public-methods


class Preisstaffel(COM):
    """
    Gibt die Staffelgrenzen der jeweiligen Preise an

    .. raw:: html

        <object data="../_static/images/bo4e/com/Preisstaffel.svg" type="image/svg+xml"></object>

    .. HINT::
        `Preisstaffel JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Preisstaffel.json>`_

    """

    # required attributes
    #: Preis pro abgerechneter Mengeneinheit
    einheitspreis: Decimal
    #: Inklusiver unterer Wert, ab dem die Staffel gilt
    staffelgrenze_von: Decimal
    #: Exklusiver oberer Wert, bis zu dem die Staffel gilt
    staffelgrenze_bis: Decimal

    # optional attributes
    #: Parameter zur Berechnung des Preises anhand der Jahresmenge und weiterer netzbezogener Parameter
    sigmoidparameter: Optional[Sigmoidparameter] = None



"""
Contains Regionskriterium class and corresponding marshmallow schema for de-/serialization
"""

from bo4e.com.com import COM
from bo4e.enum.gueltigkeitstyp import Gueltigkeitstyp
from bo4e.enum.regionskriteriumtyp import Regionskriteriumtyp

# pylint: disable=too-few-public-methods


class Regionskriterium(COM):
    """
    Komponente zur Abbildung eines Regionskriteriums

    .. raw:: html

        <object data="../_static/images/bo4e/com/Regionskriterium.svg" type="image/svg+xml"></object>

    .. HINT::
        `Regionskriterium JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Regionskriterium.json>`_

    """

    # required attributes
    gueltigkeitstyp: Gueltigkeitstyp  #: Hier wird festgelegt, ob es sich um ein einschließendes oder ausschließendes Kriterium handelt.
    regionskriteriumtyp: Regionskriteriumtyp  #: Hier wird das Kriterium selbst angegeben, z.B. Bundesland.
    wert: str
    """
    Der Wert, den das Kriterium annehmen kann, z.B. NRW.
    Im Falle des Regionskriteriumstyp BUNDESWEIT spielt dieser Wert keine Rolle.
    """



from decimal import Decimal

import pytest
from pydantic import ValidationError

from bo4e.com.tarifpreis import Tarifpreis
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.preisstatus import Preisstatus
from bo4e.enum.preistyp import Preistyp
from bo4e.enum.waehrungseinheit import Waehrungseinheit

example_tarifpreis = Tarifpreis(
    wert=Decimal(12.5),
    einheit=Waehrungseinheit.EUR,
    bezugswert=Mengeneinheit.KWH,
    preistyp=Preistyp.ARBEITSPREIS_HT,
)


class TestTarifpreis:
    def test_tarifpreis_only_required(self) -> None:
        """
        Test de-/serialisation of Tarifpreis (only has required attributes).
        """
        tarifpreis = example_tarifpreis

        json_string = tarifpreis.json(by_alias=True, ensure_ascii=False)

        assert "ARBEITSPREIS_HT" in json_string
        assert "null" in json_string

        tarifpreis_deserialized = Tarifpreis.parse_raw(json_string)

        assert isinstance(tarifpreis_deserialized.wert, Decimal)
        assert isinstance(tarifpreis_deserialized.einheit, Waehrungseinheit)
        assert isinstance(tarifpreis_deserialized.bezugswert, Mengeneinheit)
        assert isinstance(tarifpreis_deserialized.preistyp, Preistyp)
        assert tarifpreis_deserialized.beschreibung is None
        assert tarifpreis == tarifpreis_deserialized

    def test_wrong_datatype(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Tarifpreis(
                wert="blubb",  # type: ignore[arg-type]
                einheit=Waehrungseinheit.EUR,
                bezugswert=Mengeneinheit.KWH,
                preistyp=Preistyp.ARBEITSPREIS_HT,
            )

        assert "1 validation error" in str(excinfo.value)
        assert "wert" in str(excinfo.value)
        assert "value is not a valid decimal" in str(excinfo.value)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Tarifpreis(  # type: ignore[call-arg]
                wert=Decimal(3.50),
                einheit=Waehrungseinheit.EUR,
                status=Preisstatus.ENDGUELTIG,
                bezugswert=Mengeneinheit.KWH,
            )

        assert "1 validation error" in str(excinfo.value)

    def test_optional_attribute(self) -> None:
        tarifpreis = Tarifpreis(
            wert=Decimal(3.50),
            einheit=Waehrungseinheit.EUR,
            bezugswert=Mengeneinheit.KWH,
            status=Preisstatus.ENDGUELTIG,
            preistyp=Preistyp.ARBEITSPREIS_HT,
            beschreibung="Das ist ein HT Arbeitspreis",
        )

        json_string = tarifpreis.json(by_alias=True, ensure_ascii=False)

        assert "Das ist ein HT Arbeitspreis" in json_string

        tarifpreis_deserialized = Tarifpreis.parse_raw(json_string)

        assert isinstance(tarifpreis_deserialized.beschreibung, str)
        assert tarifpreis_deserialized.beschreibung == "Das ist ein HT Arbeitspreis"
        assert tarifpreis_deserialized == tarifpreis



"""
Contains Adresse class
and corresponding marshmallow schema for de-/serialization
"""
from typing import Any, Dict, Optional

from pydantic import validator

from bo4e.com.com import COM
from bo4e.enum.landescode import Landescode

# pylint: disable=too-many-instance-attributes, too-few-public-methods


class Adresse(COM):
    """
    Contains an address that can be used for most purposes.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Adresse.svg" type="image/svg+xml"></object>

    .. HINT::
        `Adresse JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Adresse.json>`_

    """

    # required attributes
    #: Die Postleitzahl; z.B: "41836"
    postleitzahl: str
    #: Bezeichnung der Stadt; z.B. "Hückelhoven"
    ort: str

    # optional attributes
    #: Bezeichnung der Straße; z.B. "Weserstraße"
    strasse: Optional[str] = None
    #: Hausnummer inkl. Zusatz; z.B. "3", "4a"
    hausnummer: Optional[str] = None
    #: Im Falle einer Postfachadresse das Postfach; Damit werden Straße und Hausnummer nicht berücksichtigt
    postfach: Optional[str] = None
    #: Zusatzhinweis zum Auffinden der Adresse, z.B. "3. Stock linke Wohnung"
    adresszusatz: Optional[str] = None
    #: Im Falle einer c/o-Adresse steht in diesem Attribut die Anrede. Z.B. "c/o Veronica Hauptmieterin"
    co_ergaenzung: Optional[str] = None
    #: Offizieller ISO-Landescode
    landescode: Landescode = Landescode.DE  # type:ignore

    # pylint: disable=no-self-argument
    @validator("postfach", always=True)
    def strasse_xor_postfach(cls, postfach: Optional[str], values: Dict[str, Any]) -> Optional[str]:
        """
        An address is valid if it contains a postfach XOR (a strasse AND hausnummer).
        This functions checks for these conditions of a valid address.

        Nur folgende Angabekombinationen sind (nach der Abfrage) möglich:
        Straße           w   f   f
        Hausnummer       w   f   f
        Postfach         f   w   f
        Postleitzahl     w   w   w
        Ort              w   w   w
        """
        if (
            values["strasse"]
            and values["hausnummer"]
            and not postfach
            or not values["strasse"]
            and not values["hausnummer"]
        ):
            return postfach
        raise ValueError('You have to define either "strasse" and "hausnummer" or "postfach".')



from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.positionsaufabschlag import PositionsAufAbschlag
from bo4e.enum.aufabschlagstyp import AufAbschlagstyp
from bo4e.enum.waehrungseinheit import Waehrungseinheit
from tests.serialization_helper import assert_serialization_roundtrip


class TestPositionsAufAbschlag:
    @pytest.mark.parametrize(
        "positionsaufabschlag, expected_json_dict",
        [
            pytest.param(
                PositionsAufAbschlag(
                    bezeichnung="foo",
                    beschreibung="bar",
                    auf_abschlagstyp=AufAbschlagstyp.ABSOLUT,
                    auf_abschlagswert=Decimal(4.25),
                    auf_abschlagswaehrung=Waehrungseinheit.EUR,
                ),
                {
                    "bezeichnung": "foo",
                    "beschreibung": "bar",
                    "aufAbschlagstyp": AufAbschlagstyp.ABSOLUT,
                    "aufAbschlagswert": Decimal("4.25"),
                    "aufAbschlagswaehrung": Waehrungseinheit.EUR,
                },
            ),
        ],
    )
    def test_serialization_roundtrip(
        self, positionsaufabschlag: PositionsAufAbschlag, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of PositionsAufAbschlag
        """
        assert_serialization_roundtrip(positionsaufabschlag, expected_json_dict)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = PositionsAufAbschlag()  # type: ignore[call-arg]

        assert "4 validation errors" in str(excinfo.value)



import pytest
from pydantic import ValidationError

from bo4e.com.zeitintervall import Zeitintervall
from bo4e.enum.zeiteinheit import Zeiteinheit


class TestZeitintervall:
    def test_zeitintervall_daten(self) -> None:
        """
        Test de-/serialisation of Zeitintervall (only has optional attributes) with option startdatum and enddatum.
        """
        zeitintervall = Zeitintervall(wert=2, zeiteinheit=Zeiteinheit.VIERTEL_STUNDE)

        json_string = zeitintervall.json(by_alias=True, ensure_ascii=False)

        assert "2" in json_string
        assert "VIERTEL_STUNDE" in json_string

        zeitintervall_deserialized = Zeitintervall.parse_raw(json_string)

        assert isinstance(zeitintervall_deserialized.wert, int)
        assert zeitintervall_deserialized.wert == 2
        assert isinstance(zeitintervall_deserialized.zeiteinheit, Zeiteinheit)
        assert zeitintervall_deserialized.zeiteinheit == Zeiteinheit.VIERTEL_STUNDE

    def test_wrong_datatype(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Zeitintervall(wert="errrrrror", zeiteinheit=Zeiteinheit.TAG)  # type: ignore[arg-type]

        assert "1 validation error" in str(excinfo.value)
        assert "wert" in str(excinfo.value)
        assert "value is not a valid integer" in str(excinfo.value)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Zeitintervall(wert=3)  # type: ignore[call-arg]

        assert "1 validation error" in str(excinfo.value)



from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.kriteriumwert import KriteriumWert
from bo4e.com.regionalegueltigkeit import RegionaleGueltigkeit
from bo4e.com.regionalepreisgarantie import RegionalePreisgarantie
from bo4e.com.regionalepreisstaffel import RegionalePreisstaffel
from bo4e.enum.gueltigkeitstyp import Gueltigkeitstyp
from bo4e.enum.tarifregionskriterium import Tarifregionskriterium
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_sigmoidparameter import example_sigmoidparameter

example_regionale_preisstaffel = RegionalePreisstaffel(
    einheitspreis=Decimal(40.0),
    staffelgrenze_von=Decimal(12.5),
    staffelgrenze_bis=Decimal(25.0),
    sigmoidparameter=example_sigmoidparameter,
    regionale_gueltigkeit=RegionaleGueltigkeit(
        gueltigkeitstyp=Gueltigkeitstyp.NUR_IN,
        kriteriums_werte=[KriteriumWert(kriterium=Tarifregionskriterium.POSTLEITZAHL, wert="01069")],
    ),
)


class TestRegionalePreisstaffel:
    @pytest.mark.parametrize(
        "regionale_preisstaffel, expected_json_dict",
        [
            pytest.param(
                example_regionale_preisstaffel,
                {
                    "regionaleGueltigkeit": {
                        "gueltigkeitstyp": "NUR_IN",
                        "kriteriumsWerte": [{"kriterium": Tarifregionskriterium.POSTLEITZAHL, "wert": "01069"}],
                    },
                    "einheitspreis": Decimal("40"),
                    "sigmoidparameter": {"A": Decimal("1"), "B": Decimal("2"), "C": Decimal("3"), "D": Decimal("4")},
                    "staffelgrenzeVon": Decimal("12.5"),
                    "staffelgrenzeBis": Decimal("25"),
                },
                id="maximal attributes"
                # the messing sigmoidparameter is tested in the Preisstaffel tests
            ),
        ],
    )
    def test_serialization_roundtrip(
        self, regionale_preisstaffel: RegionalePreisstaffel, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of RegionalePreisgarantie with maximal attributes.
        """
        assert_serialization_roundtrip(regionale_preisstaffel, expected_json_dict)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = RegionalePreisgarantie()  # type: ignore[call-arg]

        assert "3 validation errors" in str(excinfo.value)



from decimal import Decimal

import pytest
from pydantic import ValidationError

from bo4e.com.preisposition import Preisposition
from bo4e.enum.bdewartikelnummer import BDEWArtikelnummer
from bo4e.enum.bemessungsgroesse import Bemessungsgroesse
from bo4e.enum.kalkulationsmethode import Kalkulationsmethode
from bo4e.enum.leistungstyp import Leistungstyp
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.tarifzeit import Tarifzeit
from bo4e.enum.waehrungseinheit import Waehrungseinheit
from bo4e.enum.zeiteinheit import Zeiteinheit
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_preisstaffel import example_preisstaffel

example_preisposition = Preisposition(
    berechnungsmethode=Kalkulationsmethode.AP_TRANSPORT_ODER_VERTEILNETZ,
    leistungstyp=Leistungstyp.SONSTIGER_PREIS,
    leistungsbezeichnung="Foo",
    preiseinheit=Waehrungseinheit.EUR,
    bezugsgroesse=Mengeneinheit.KVARH,
    zeitbasis=Zeiteinheit.HALBJAHR,
    tarifzeit=Tarifzeit.TZ_HT,
    bdew_artikelnummer=BDEWArtikelnummer.AUSGLEICHSENERGIE_UNTERDECKUNG,
    zonungsgroesse=Bemessungsgroesse.BENUTZUNGSDAUER,
    freimenge_blindarbeit=Decimal(50),  # %
    freimenge_leistungsfaktor=Decimal(1.0),
    preisstaffeln=[example_preisstaffel],
    gruppenartikel_id="1-2-3",
)


class TestPreisposition:
    @pytest.mark.parametrize(
        "preisposition",
        [
            pytest.param(example_preisposition),
        ],
    )
    def test_serialization_roundtrip(self, preisposition: Preisposition) -> None:
        """
        Test de-/serialisation of Preisposition
        """
        assert_serialization_roundtrip(preisposition)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Preisposition()  # type: ignore[call-arg]
        assert "6 validation errors" in str(excinfo.value)



# pylint:disable=missing-module-docstring

from bo4e.enum.strenum import StrEnum


class Geschaeftspartnerrolle(StrEnum):
    """
    Diese Rollen kann ein Geschäftspartner einnehmen.
    """

    LIEFERANT = "LIEFERANT"  #: Lieferant
    DIENSTLEISTER = "DIENSTLEISTER"  #: Dienstleister
    KUNDE = "KUNDE"  #: Kunde
    INTERESSENT = "INTERESSENT"  #: Interessent
    MARKTPARTNER = "MARKTPARTNER"  #: Marktpartner



"""
Contains base class for all components
"""

from decimal import Decimal
from typing import Type, TypeVar

from humps.main import camelize

# pylint: disable=no-name-in-module
from pydantic import BaseModel


# pylint: disable=too-few-public-methods
#
class COM(BaseModel):
    """
    base class for all components

    .. raw:: html

        <object data="../_static/images/bo4e/com/COM.svg" type="image/svg+xml"></object>

    .. HINT::
        `COM JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/COM.json>`_

    """

    class Config:
        """
        basic configuration for pydantic's behaviour
        """

        alias_generator = camelize
        allow_population_by_field_name = True
        json_encoders = {Decimal: str}


# pylint: disable=invalid-name
#: Any type derived from COM including those that do not directly inherit from COM
TCom = TypeVar("TCom", bound=Type[COM])
# todo: find out if this way of typing is correct



# pylint: disable=missing-module-docstring
from bo4e.enum.strenum import StrEnum


class Angebotsstatus(StrEnum):
    """
    Gibt den Status eines Angebotes an.
    """

    KONZEPTION = "KONZEPTION"  #: Konzeption
    UNVERBINDLICH = "UNVERBINDLICH"  #: unverbindlich
    VERBINDLICH = "VERBINDLICH"  #: verbindlich
    BEAUFTRAGT = "BEAUFTRAGT"  #: beauftragt
    UNGUELTIG = "UNGUELTIG"  #: ungültig
    ABGELEHNT = "ABGELEHNT"  #: abgelehnt
    NACHGEFASST = "NACHGEFASST"  #: nachgefasst
    AUSSTEHEND = "AUSSTEHEND"  #: ausstehend
    ERLEDIGT = "ERLEDIGT"  #: erledigt



from decimal import Decimal
from typing import Any, Dict

import pytest
from pydantic import ValidationError

from bo4e.com.ausschreibungsdetail import Ausschreibungsdetail
from bo4e.com.menge import Menge
from bo4e.enum.landescode import Landescode
from bo4e.enum.mengeneinheit import Mengeneinheit
from bo4e.enum.netzebene import Netzebene
from bo4e.enum.zaehlertyp import Zaehlertyp
from bo4e.enum.zeiteinheit import Zeiteinheit
from tests.serialization_helper import assert_serialization_roundtrip
from tests.test_adresse import example_adresse
from tests.test_menge import example_menge
from tests.test_zeitraum import example_zeitraum

example_ausschreibungsdetail = Ausschreibungsdetail(
    marktlokations_id="56789012345",
    netzebene_lieferung=Netzebene.MSP,
    netzebene_messung=Netzebene.NSP,
    marktlokationsadresse=example_adresse,
    lieferzeitraum=example_zeitraum,
    rechnungsadresse=example_adresse,
)
example_ausschreibungsdetail_dict = {
    "zaehlernummer": None,
    "zaehlertechnik": None,
    "kunde": None,
    "marktlokationsbezeichnung": None,
    "lieferzeitraum": {
        "dauer": Decimal("5"),
        "startdatum": None,
        "endzeitpunkt": None,
        "enddatum": None,
        "einheit": Zeiteinheit.TAG,
        "startzeitpunkt": None,
    },
    "marktlokationsadresse": {
        "landescode": Landescode.DE,  # type: ignore[attr-defined]
        "hausnummer": "27A",
        "strasse": "Nördliche Münchner Straße",
        "postleitzahl": "82031",
        "ort": "Grünwald",
        "adresszusatz": None,
        "postfach": None,
        "coErgaenzung": None,
    },
    "rechnungsadresse": {
        "landescode": Landescode.DE,  # type: ignore[attr-defined]
        "hausnummer": "27A",
        "strasse": "Nördliche Münchner Straße",
        "postleitzahl": "82031",
        "ort": "Grünwald",
        "adresszusatz": None,
        "postfach": None,
        "coErgaenzung": None,
    },
    "netzbetreiber": None,
    "netzebeneLieferung": Netzebene.MSP,
    "prognoseArbeitLieferzeitraum": None,
    "netzebeneMessung": Netzebene.NSP,
    "prognoseLeistung": None,
    "lastgangVorhanden": None,
    "prognoseJahresarbeit": None,
    "marktlokationsId": "56789012345",
}


class TestAusschreibungsdetail:
    @pytest.mark.parametrize(
        "ausschreibungsdetail, expected_json_dict",
        [
            pytest.param(
                Ausschreibungsdetail(
                    marktlokations_id="56789012345",
                    netzebene_lieferung=Netzebene.MSP,
                    netzebene_messung=Netzebene.NSP,
                    marktlokationsadresse=example_adresse,
                    lieferzeitraum=example_zeitraum,
                    netzbetreiber="Stromnetz Hamburg GmbH",
                    kunde="Dei Mudder ihr Kunde",
                    zaehlernummer="1YSK4234092304",
                    marktlokationsbezeichnung="Zentraler Einkauf, Hamburg",
                    zaehlertechnik=Zaehlertyp.LEISTUNGSZAEHLER,
                    lastgang_vorhanden=True,
                    prognose_leistung=Menge(wert=Decimal(40), einheit=Mengeneinheit.KW),
                    prognose_arbeit_lieferzeitraum=Menge(wert=Decimal(2500), einheit=Mengeneinheit.KWH),
                    prognose_jahresarbeit=Menge(wert=Decimal(2500), einheit=Mengeneinheit.KWH),
                    rechnungsadresse=example_adresse,
                ),
                {
                    "netzbetreiber": "Stromnetz Hamburg GmbH",
                    "lieferzeitraum": {
                        "enddatum": None,
                        "startdatum": None,
                        "einheit": Zeiteinheit.TAG,
                        "endzeitpunkt": None,
                        "dauer": Decimal("5"),
                        "startzeitpunkt": None,
                    },
                    "zaehlertechnik": Zaehlertyp.LEISTUNGSZAEHLER,
                    "kunde": "Dei Mudder ihr Kunde",
                    "marktlokationsbezeichnung": "Zentraler Einkauf, Hamburg",
                    "marktlokationsadresse": {
                        "hausnummer": "27A",
                        "adresszusatz": None,
                        "postfach": None,
                        "postleitzahl": "82031",
                        "landescode": Landescode.DE,  # type: ignore[attr-defined]
                        "ort": "Grünwald",
                        "strasse": "Nördliche Münchner Straße",
                        "coErgaenzung": None,
                    },
                    "rechnungsadresse": {
                        "hausnummer": "27A",
                        "adresszusatz": None,
                        "postfach": None,
                        "postleitzahl": "82031",
                        "landescode": Landescode.DE,  # type: ignore[attr-defined]
                        "ort": "Grünwald",
                        "strasse": "Nördliche Münchner Straße",
                        "coErgaenzung": None,
                    },
                    "zaehlernummer": "1YSK4234092304",
                    "prognoseJahresarbeit": {"wert": Decimal("2500"), "einheit": Mengeneinheit.KWH},
                    "netzebeneLieferung": Netzebene.MSP,
                    "marktlokationsId": "56789012345",
                    "prognoseLeistung": {"wert": Decimal("40"), "einheit": Mengeneinheit.KW},
                    "lastgangVorhanden": True,
                    "netzebeneMessung": Netzebene.NSP,
                    "prognoseArbeitLieferzeitraum": {"wert": Decimal("2500"), "einheit": Mengeneinheit.KWH},
                },
            ),
            pytest.param(example_ausschreibungsdetail, example_ausschreibungsdetail_dict),
        ],
    )
    def test_serialization_roundtrip(
        self, ausschreibungsdetail: Ausschreibungsdetail, expected_json_dict: Dict[str, Any]
    ) -> None:
        """
        Test de-/serialisation of Ausschreibungsdetail
        """
        assert_serialization_roundtrip(ausschreibungsdetail, expected_json_dict)

    def test_missing_required_attribute(self) -> None:
        with pytest.raises(ValidationError) as excinfo:
            _ = Ausschreibungsdetail()  # type: ignore[call-arg]

        assert "5 validation errors" in str(excinfo.value)
        # 'lokations_id', 'netzebene_lieferung', 'netzebene_messung', 'lokationsadresse', and 'lieferzeitraum'



"""
Contains Dienstleistung class
and corresponding marshmallow schema for de-/serialization
"""

from bo4e.com.com import COM
from bo4e.enum.dienstleistungstyp import Dienstleistungstyp

# pylint: disable=too-few-public-methods


class Dienstleistung(COM):
    """
    Abbildung einer abrechenbaren Dienstleistung.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Dienstleistung.svg" type="image/svg+xml"></object>

    .. HINT::
        `Dienstleistung JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/Hochfrequenz/BO4E-python/main/json_schemas/com/Dienstleistung.json>`_

    """

    # required attributes
    #: Kennzeichnung der Dienstleistung
    dienstleistungstyp: Dienstleistungstyp
    #: Bezeichnung der Dienstleistung
    bezeichnung: str
